import datetime
import logging
import os
import threading
from threading import RLock

from time import time, sleep

from geometry_msgs.msg import Point
from typing_extensions import List, Dict, Tuple, Optional, Callable, Union

from .socket import MultiverseSocket, MultiverseMetaData
from ...config.multiverse_conf import MultiverseConfig as Conf
from ...datastructures.dataclasses import MultiverseRayResult, MultiverseContactPoint, \
    AxisAlignedBoundingBox
from ...datastructures.enums import (MultiverseAPIName as API, MultiverseBodyProperty as BodyProperty,
                                     MultiverseProperty as Property)
from ...datastructures.pose import PoseStamped
from ...failures import MultiverseFailedAPIResponse
from ...utils import wxyz_to_xyzw
from ...world_concepts.constraints import Constraint
from ...world_concepts.world_object import Object, Link


class MultiverseClient(MultiverseSocket):

    def __init__(self, name: str, port: int, is_prospection_world: bool = False,
                 simulation_wait_time_factor: float = 1.0, **kwargs):
        """
        Initialize the Multiverse client, which connects to the Multiverse server.

        :param name: The name of the client.
        :param port: The port of the client.
        :param is_prospection_world: Whether the client is connected to the prospection world.
        :param simulation_wait_time_factor: The simulation wait time factor (default is 1.0), which can be used to
            increase or decrease the wait time for the simulation.
        """
        meta_data = MultiverseMetaData()
        meta_data.simulation_name = ((Conf.prospection_world_prefix if is_prospection_world else "belief_state")
                                     + "_" + name)
        meta_data.world_name = Conf.prospection_world_prefix if is_prospection_world else "belief_state"
        self.is_prospection_world = is_prospection_world
        super().__init__(port=str(port), meta_data=meta_data)
        self.simulation_wait_time_factor = simulation_wait_time_factor
        self.run()


class MultiverseReader(MultiverseClient):
    MAX_WAIT_TIME_FOR_DATA: datetime.timedelta = Conf.READER_MAX_WAIT_TIME_FOR_DATA
    """
    The maximum wait time for the data in seconds.
    """

    def __init__(self, name: str, port: int, is_prospection_world: bool = False,
                 simulation_wait_time_factor: float = 1.0, **kwargs):
        """
        Initialize the Multiverse reader, which reads the data from the Multiverse server in a separate thread.
        This class provides methods to get data (e.g., position, orientation) from the Multiverse server.

        :param port: The port of the Multiverse reader client.
        :param is_prospection_world: Whether the reader is connected to the prospection world.
        :param simulation_wait_time_factor: The simulation wait time factor.
        """
        super().__init__(name, port, is_prospection_world, simulation_wait_time_factor=simulation_wait_time_factor)

        self.request_meta_data["receive"][""] = [""]

        self.data_lock = threading.Lock()
        self.thread = threading.Thread(target=self.receive_all_data_from_server)
        self.stop_thread = False

        self.thread.start()

    def get_body_pose(self, name: str, wait: bool = False) -> Optional[Dict[str, List[float]]]:
        """
        Get the body pose from the multiverse server.

        :param name: The name of the body.
        :param wait: Whether to wait for the data.
        :return: The position and orientation of the body.
        """
        return self.get_body_data(name, [BodyProperty.POSITION, BodyProperty.ORIENTATION], wait=wait)

    def get_multiple_body_poses(self, body_names: List[str], wait: bool = False) -> Optional[Dict[str, PoseStamped]]:
        """
        Get the body poses from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param wait: Whether to wait for the data.
        :return: The positions and orientations of the bodies as a dictionary.
        """
        data = self.get_multiple_body_data(body_names,
                                           {name: [BodyProperty.POSITION, BodyProperty.ORIENTATION]
                                            for name in body_names
                                            },
                                           wait=wait)
        if data is not None:
            return {name: PoseStamped(data[name][BodyProperty.POSITION.value],
                                      wxyz_to_xyzw(data[name][BodyProperty.ORIENTATION.value]))
                    for name in body_names}

    def get_body_position(self, name: str, wait: bool = False) -> Optional[List[float]]:
        """
        Get the body position from the multiverse server.

        :param name: The name of the body.
        :param wait: Whether to wait for the data.
        :return: The position of the body.
        """
        return self.get_body_property(name, BodyProperty.POSITION, wait=wait)

    def get_multiple_body_positions(self, body_names: List[str],
                                    wait: bool = False) -> Optional[Dict[str, List[float]]]:
        """
        Get the body positions from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param wait: Whether to wait for the data.
        :return: The positions of the bodies as a dictionary.
        """
        return self.get_multiple_body_properties(body_names, [BodyProperty.POSITION], wait=wait)

    def get_body_orientation(self, name: str, wait: bool = False) -> Optional[List[float]]:
        """
        Get the body orientation from the multiverse server.

        :param name: The name of the body.
        :param wait: Whether to wait for the data.
        :return: The orientation of the body.
        """
        return self.get_body_property(name, BodyProperty.ORIENTATION, wait=wait)

    def get_multiple_body_orientations(self, body_names: List[str],
                                       wait: bool = False) -> Optional[Dict[str, List[float]]]:
        """
        Get the body orientations from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param wait: Whether to wait for the data.
        :return: The orientations of the bodies as a dictionary.
        """
        data = self.get_multiple_body_properties(body_names, [BodyProperty.ORIENTATION], wait=wait)
        if data is not None:
            return {name: wxyz_to_xyzw(data[name][BodyProperty.ORIENTATION.value]) for name in body_names}

    def get_body_property(self, name: str, property_: Property, wait: bool = False) -> Optional[List[float]]:
        """
        Get the body property from the multiverse server.

        :param name: The name of the body.
        :param property_: The property of the body as a Property.
        :param wait: Whether to wait for the data.
        :return: The property of the body.
        """
        data = self.get_body_data(name, [property_], wait=wait)
        if data is not None:
            return data[property_.value]

    def get_multiple_body_properties(self, body_names: List[str], properties: List[Property],
                                     wait: bool = False) -> Optional[Dict[str, Dict[str, List[float]]]]:
        """
        Get the body properties from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param properties: The properties of the bodies.
        :param wait: Whether to wait for the data.
        :return: The properties of the bodies as a dictionary.
        """
        return self.get_multiple_body_data(body_names, {name: properties for name in body_names}, wait=wait)

    def get_body_data(self, name: str,
                      properties: Optional[List[Property]] = None,
                      wait: bool = False) -> Optional[Dict]:
        """
        Get the body data from the multiverse server.

        :param name: The name of the body.
        :param properties: The properties of the body.
        :param wait: Whether to wait for the data.
        :return: The body data as a dictionary.
        """
        if wait:
            return self.wait_for_body_data(name, properties)

        data = self.get_received_data()
        if self.check_for_body_data(name, data, properties):
            return data[name]

    def get_multiple_body_data(self, body_names: List[str],
                               properties: Optional[Dict[str, List[Property]]] = None,
                               wait: bool = False) -> Optional[Dict]:
        """
        Get the body data from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param properties: The properties of the bodies.
        :param wait: Whether to wait for the data.
        :return: The body data as a dictionary.
        """

        if wait:
            return self.wait_for_multiple_body_data(body_names, properties)

        data = self.get_received_data()
        if self.check_multiple_body_data(body_names, data, properties):
            return {name: data[name] for name in body_names}

    def wait_for_body_data(self, name: str, properties: Optional[List[Property]] = None) -> Dict:
        """
        Wait for the body data from the multiverse server.

        :param name: The name of the body.
        :param properties: The properties of the body.
        :return: The body data as a dictionary.
        """
        return self._wait_for_body_data_template(name, self.check_for_body_data, properties)[name]

    def wait_for_multiple_body_data(self, body_names: List[str],
                                    properties: Optional[Dict[str, List[Property]]] = None) -> Dict:
        """
        Wait for the body data from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param properties: The properties of the bodies.
        :return: The body data as a dictionary.
        """
        return self._wait_for_body_data_template(body_names, self.check_multiple_body_data, properties)

    def _wait_for_body_data_template(self, body_names: Union[str, List[str]],
                                     check_func: Callable[[Union[str, List[str]], Dict, Union[Dict, List]], bool],
                                     properties: Optional[Union[Dict, List]] = None) -> Dict:
        """
        Wait for the body data from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param properties: The properties of the bodies.
        :param check_func: The function to check if the data is received.
        :return: The body data as a dictionary.
        """
        start = time()
        data_received_flag = False
        while time() - start < self.MAX_WAIT_TIME_FOR_DATA.total_seconds():
            received_data = self.get_received_data()
            data_received_flag = check_func(body_names, received_data, properties)
            if data_received_flag:
                return received_data
        if not data_received_flag:
            properties_str = "Data" if properties is None else f"Properties {properties}"
            msg = f"{properties_str} for {body_names} not received within {self.MAX_WAIT_TIME_FOR_DATA} seconds"
            logging.error(msg)
            raise ValueError(msg)

    def check_multiple_body_data(self, body_names: List[str], data: Dict,
                                 properties: Optional[Dict[str, List[Property]]] = None) -> bool:
        """
        Check if the body data is received from the multiverse server for multiple bodies.

        :param body_names: The names of the bodies.
        :param data: The data received from the multiverse server.
        :param properties: The properties of the bodies.
        :return: Whether the body data is received.
        """
        if properties is None:
            return all([self.check_for_body_data(name, data) for name in body_names])
        else:
            return all([self.check_for_body_data(name, data, properties[name]) for name in body_names])

    @staticmethod
    def check_for_body_data(name: str, data: Dict, properties: Optional[List[Property]] = None) -> bool:
        """
        Check if the body data is received from the multiverse server.

        :param name: The name of the body.
        :param data: The data received from the multiverse server.
        :param properties: The properties of the body.
        :return: Whether the body data is received.
        """
        if properties is None:
            return name in data
        else:
            return name in data and all([prop.value in data[name] and None not in data[name][prop.value]
                                         for prop in properties])

    def get_received_data(self):
        """
        Get the latest received data from the multiverse server.
        """
        self.data_lock.acquire()
        data = self.response_meta_data["receive"]
        self.data_lock.release()
        return data

    def receive_all_data_from_server(self):
        """
        Get all data from the multiverse server.
        """
        while not self.stop_thread:
            self.request_meta_data["receive"][""] = [""]
            self.data_lock.acquire()
            self.send_and_receive_meta_data()
            self.data_lock.release()
            sleep(0.01)
        self.stop()

    def join(self):
        self.thread.join()


class MultiverseWriter(MultiverseClient):

    def __init__(self, name: str, port: int, simulation: Optional[str] = None,
                 is_prospection_world: bool = False,
                 simulation_wait_time_factor: float = 1.0, **kwargs):
        """
        Initialize the Multiverse writer, which writes the data to the Multiverse server.
        This class provides methods to send data (e.g., position, orientation) to the Multiverse server.

        :param port: The port of the Multiverse writer client.
        :param simulation: The name of the simulation that the writer is connected to
         (usually the name defined in the .muv file).
        :param is_prospection_world: Whether the writer is connected to the prospection world.
        :param simulation_wait_time_factor: The wait time factor for the simulation (default is 1.0), which can be used
         to increase or decrease the wait time for the simulation.
        """
        super().__init__(name, port, is_prospection_world, simulation_wait_time_factor=simulation_wait_time_factor)
        self.simulation = simulation
        self.lock = RLock()

    def spawn_robot_with_actuators(self, robot_name: str,
                                   actuator_joint_commands: Optional[Dict[str, List[str]]] = None) -> None:
        """
        Spawn the robot with controlled actuators in the simulation.

        :param robot_name: The name of the robot.
        :param actuator_joint_commands: A dictionary mapping actuator names to joint command names.
        """
        send_meta_data = {robot_name: [BodyProperty.POSITION.value, BodyProperty.ORIENTATION.value,
                                       BodyProperty.RELATIVE_VELOCITY.value]}
        relative_velocity = [0.0] * 6
        data = [self.sim_time] + [0.0] * 3 + [1, 0, 0, 0] + relative_velocity
        self.send_data_to_server(data, send_meta_data=send_meta_data, receive_meta_data=actuator_joint_commands)

    def _reset_request_meta_data(self, set_simulation_name: bool = True):
        """
        Reset the request metadata.

        :param set_simulation_name: Whether to set the simulation name to the value of self.simulation_name.
        """
        self.request_meta_data = {
            "meta_data": self._meta_data.__dict__.copy(),
            "send": {},
            "receive": {},
        }
        if self.simulation is not None and set_simulation_name:
            self.request_meta_data["meta_data"]["simulation_name"] = self.simulation

    def set_body_pose(self, body_name: str, position: List[float], orientation: List[float]) -> None:
        """
        Set the body pose in the simulation.

        :param body_name: The name of the body.
        :param position: The position of the body.
        :param orientation: The orientation of the body.
        """
        self.send_body_data_to_server(body_name,
                                      {BodyProperty.POSITION: position,
                                       BodyProperty.ORIENTATION: orientation,
                                       BodyProperty.RELATIVE_VELOCITY: [0.0] * 6})

    def set_multiple_body_poses(self, body_data: Dict[str, Dict[BodyProperty, List[float]]]) -> None:
        """
        Set the body poses in the simulation for multiple bodies.

        :param body_data: The data to be sent for multiple bodies.
        """
        self.send_multiple_body_data_to_server(body_data)

    def set_body_position(self, body_name: str, position: List[float]) -> None:
        """
        Set the body position in the simulation.

        :param body_name: The name of the body.
        :param position: The position of the body.
        """
        self.set_body_property(body_name, BodyProperty.POSITION, position)

    def set_body_orientation(self, body_name: str, orientation: List[float]) -> None:
        """
        Set the body orientation in the simulation.

        :param body_name: The name of the body.
        :param orientation: The orientation of the body.
        """
        self.set_body_property(body_name, BodyProperty.ORIENTATION, orientation)

    def set_body_property(self, body_name: str, property_: Property, value: List[float]) -> None:
        """
        Set the body property in the simulation.

        :param body_name: The name of the body.
        :param property_: The property of the body.
        :param value: The value of the property.
        """
        self.send_body_data_to_server(body_name, {property_: value})

    def remove_body(self, body_name: str) -> None:
        """
        Remove the body from the simulation.

        :param body_name: The name of the body.
        """
        self.send_data_to_server([self.sim_time],
                                 send_meta_data={body_name: []},
                                 receive_meta_data={body_name: []})

    def reset_world(self) -> None:
        """
        Reset the world in the simulation.
        """
        self.send_data_to_server([0], set_simulation_name=False)

    def send_body_data_to_server(self, body_name: str, body_data: Dict[Property, List[float]]) -> Dict:
        """
        Send data to the multiverse server.

        :param body_name: The name of the body.
        :param body_data: The data to be sent.
        :return: The response from the server.
        """
        send_meta_data = {body_name: list(map(str, body_data.keys()))}
        flattened_data = [value for data in body_data.values() for value in data]
        return self.send_data_to_server([self.sim_time, *flattened_data], send_meta_data=send_meta_data)

    def send_multiple_body_data_to_server(self, body_data: Dict[str, Dict[Property, List[float]]]) -> Dict:
        """
        Send data to the multiverse server for multiple bodies.

        :param body_data: The data to be sent for multiple bodies.
        :return: The response from the server.
        """
        send_meta_data = {body_name: list(map(str, data.keys())) for body_name, data in body_data.items()}
        response_meta_data = self.send_meta_data_and_get_response(send_meta_data)
        body_names = list(response_meta_data["send"].keys())
        flattened_data = [value for body_name in body_names for data in body_data[body_name].values()
                          for value in data]
        self.lock.acquire()
        self.send_data = [self.sim_time, *flattened_data]
        self.send_and_receive_data()
        response_meta_data = self.response_meta_data
        self.lock.release()
        return response_meta_data

    def send_meta_data_and_get_response(self, send_meta_data: Dict) -> Dict:
        """
        Send metadata to the multiverse server and get the response.

        :param send_meta_data: The metadata to be sent.
        :return: The response from the server.
        """
        self.lock.acquire()
        self._reset_request_meta_data()
        self.request_meta_data["send"] = send_meta_data
        self.send_and_receive_meta_data()
        response_meta_data = self.response_meta_data
        self.lock.release()
        return response_meta_data

    def send_data_to_server(self, data: List,
                            send_meta_data: Optional[Dict] = None,
                            receive_meta_data: Optional[Dict] = None,
                            set_simulation_name: bool = True) -> Dict:
        """
        Send data to the multiverse server.

        :param data: The data to be sent.
        :param send_meta_data: The metadata to be sent.
        :param receive_meta_data: The metadata to be received.
        :param set_simulation_name: Whether to set the simulation name to the value of self.simulation.
        :return: The response from the server.
        """
        self.lock.acquire()
        self._reset_request_meta_data(set_simulation_name=set_simulation_name)
        if send_meta_data:
            self.request_meta_data["send"] = send_meta_data
        if receive_meta_data:
            self.request_meta_data["receive"] = receive_meta_data
        self.send_and_receive_meta_data()
        self.send_data = data
        self.send_and_receive_data()
        response_meta_data = self.response_meta_data
        self.lock.release()
        return response_meta_data


class MultiverseController(MultiverseWriter):

    def __init__(self, name: str, port: int, is_prospection_world: bool = False, **kwargs):
        """
        Initialize the Multiverse controller, which controls the robot in the simulation.
        This class provides methods to send controller data to the Multiverse server.

        :param port: The port of the Multiverse controller client.
        :param is_prospection_world: Whether the controller is connected to the prospection world.
        """
        super().__init__(name, port, is_prospection_world=is_prospection_world)

    def init_controller(self, actuator_joint_commands: Dict[str, List[str]]) -> None:
        """
        Initialize the controller by sending the controller data to the multiverse server.

        :param actuator_joint_commands: A dictionary mapping actuator names to joint command names.
        """
        self.send_data_to_server([self.sim_time] + [0.0] * len(actuator_joint_commands),
                                 send_meta_data=actuator_joint_commands)


class MultiverseAPI(MultiverseClient):
    API_REQUEST_WAIT_TIME: datetime.timedelta = datetime.timedelta(milliseconds=200)
    """
    The wait time for the API request in seconds.
    """
    APIs_THAT_NEED_WAIT_TIME: List[API] = [API.ATTACH, API.DETACH]

    def __init__(self, name: str, port: int, simulation: str, is_prospection_world: bool = False,
                 simulation_wait_time_factor: float = 1.0):
        """
        Initialize the Multiverse API, which sends API requests to the Multiverse server.
        This class provides methods like attach and detach objects, get contact points, and other API requests.

        :param port: The port of the Multiverse API client.
        :param simulation: The name of the simulation that the API is connected to
         (usually the name defined in the .muv file).
        :param is_prospection_world: Whether the API is connected to the prospection world.
        :param simulation_wait_time_factor: The simulation wait time factor, which can be used to increase or decrease
            the wait time for the simulation.
        """
        super().__init__(name, port, is_prospection_world, simulation_wait_time_factor=simulation_wait_time_factor)
        self.simulation = simulation
        self.wait: bool = False  # Whether to wait after sending the API request.

    def get_body_bounding_box(self, body_name: str,
                              with_children: bool = False) -> Union[AxisAlignedBoundingBox, List[AxisAlignedBoundingBox]]:
        """
        Get the body bounding box from the multiverse server, they are with respect to the body's frame.
        """
        bounding_boxes_data = self._get_bounding_box(body_name, with_children)
        bounding_boxes = []
        for bounding_box in bounding_boxes_data:
            origin = Point(x=bounding_box[0], y=bounding_box[1], z=bounding_box[2])
            size = Point(x=bounding_box[3], y=bounding_box[4], z=bounding_box[5])
            bounding_boxes.append(AxisAlignedBoundingBox.from_origin_and_half_extents(origin, size))
        if with_children:
            return bounding_boxes
        return bounding_boxes[0]

    def _get_bounding_box(self, body_name: str, with_children: bool = False) -> List[List[float]]:
        """
        Get the body bounding box from the multiverse server.
        """
        params = [body_name]
        if with_children:
            params.append("with_children")
        response = self._request_single_api_callback(API.GET_BOUNDING_BOX, *params)[0]
        return [list(map(float, bounding_box.split())) for bounding_box in response]

    def save(self, save_name: str, save_directory: Optional[str] = None) -> str:
        """
        Save the current state of the simulation.

        :param save_name: The name of the save.
        :param save_directory: The path to save the simulation, can be relative or absolute. If the path is relative,
         it will be saved in the saved folder in multiverse.
        :return: The save path.
        """
        response = self._request_single_api_callback(API.SAVE, self.get_save_path(save_name, save_directory))
        return response[0]

    def load(self, save_name: str, save_directory: Optional[str] = None) -> None:
        """
        Load the saved state of the simulation.

        :param save_name: The name of the save.
        :param save_directory: The path to load the simulation, can be relative or absolute. If the path is relative,
         it will be loaded from the saved folder in multiverse.
        """
        self._request_single_api_callback(API.LOAD, self.get_save_path(save_name, save_directory))

    @staticmethod
    def get_save_path(save_name: str, save_directory: Optional[str] = None) -> str:
        """
        Get the save path.

        :param save_name: The save name.
        :param save_directory: The save directory.
        :return: The save path.
        """
        return save_name if not save_directory else os.path.join(save_directory, save_name)

    def attach(self, constraint: Constraint) -> None:
        """
        Request to attach the child link to the parent link.

        :param constraint: The constraint.
        """
        self.wait = True
        parent_link_name, child_link_name = self.get_constraint_link_names(constraint)
        attachment_pose = self._get_attachment_pose_as_string(constraint)
        self._attach(child_link_name, parent_link_name, attachment_pose)

    def _attach(self, child_link_name: str, parent_link_name: str, attachment_pose: str) -> None:
        """
        Attach the child link to the parent link.

        :param child_link_name: The name of the child link.
        :param parent_link_name: The name of the parent link.
        :param attachment_pose: The attachment pose.
        """
        self._request_single_api_callback(API.ATTACH, child_link_name, parent_link_name,
                                          attachment_pose)

    def get_constraint_link_names(self, constraint: Constraint) -> Tuple[str, str]:
        """
        Get the link names of the constraint.

        :param constraint: The constraint.
        :return: The link names of the constraint.
        """
        return self.get_parent_link_name(constraint), self.get_constraint_child_link_name(constraint)

    def get_parent_link_name(self, constraint: Constraint) -> str:
        """
        Get the parent link name of the constraint.

        :param constraint: The constraint.
        :return: The parent link name of the constraint.
        """
        return self.get_link_name_for_constraint(constraint.parent_link)

    def get_constraint_child_link_name(self, constraint: Constraint) -> str:
        """
        Get the child link name of the constraint.

        :param constraint: The constraint.
        :return: The child link name of the constraint.
        """
        return self.get_link_name_for_constraint(constraint.child_link)

    @staticmethod
    def get_link_name_for_constraint(link: Link) -> str:
        """
        Get the link name from link object, if the link belongs to a one link object, return the object name.

        :param link: The link.
        :return: The link name.
        """
        return link.name if not link.is_only_link else link.object.name

    def detach(self, constraint: Constraint) -> None:
        """
        Request to detach the child link from the parent link.

        :param constraint: The constraint.
        """
        parent_link_name, child_link_name = self.get_constraint_link_names(constraint)
        self._detach(child_link_name, parent_link_name)

    def _detach(self, child_link_name: str, parent_link_name: str) -> None:
        """
        Detach the child link from the parent link.

        :param child_link_name: The name of the child link.
        :param parent_link_name: The name of the parent link.
        """
        self._request_single_api_callback(API.DETACH, child_link_name, parent_link_name)

    def _get_attachment_pose_as_string(self, constraint: Constraint) -> str:
        """
        Get the attachment pose as a string.

        :param constraint: The constraint.
        :return: The attachment pose as a string.
        """
        pose = constraint.parent_to_child_transform.to_pose_stamped()
        return self._pose_to_string(pose)

    @staticmethod
    def _pose_to_string(pose: PoseStamped) -> str:
        """
        Convert the pose to a string.

        :param pose: The pose.
        :return: The pose as a string.
        """
        return f"{pose.position.x} {pose.position.y} {pose.position.z} {pose.orientation.w} {pose.orientation.x} " \
               f"{pose.orientation.y} {pose.orientation.z}"

    def check_object_exists(self, obj: Object) -> bool:
        """
        Check if the object exists in the simulation.

        :param obj: The object.
        :return: Whether the object exists in the simulation.
        """
        return self._request_single_api_callback(API.EXIST, obj.name)[0] == 'yes'

    def get_resultant_force_and_torque_on_object(self, obj: Object) -> Tuple[List[float], List[float]]:
        """
        Get the resultant force and torque on the object.

        :param obj: The object.
        :return: The resultant force and torque on the object.
        """
        effort = self._request_single_api_callback(API.GET_CONSTRAINT_EFFORT, obj.name)
        return self._parse_constraint_effort(effort)

    def get_contact_points_between_bodies(self, body_1_name: str, body_2_name: str) -> List[MultiverseContactPoint]:
        """
        Request the contact points between two bodies.

        :param body_1_name: The name of the first body.
        :param body_2_name: The name of the second body.
        :return: The contact points between the bodies as a list of MultiverseContactPoint.
        """
        points = self._request_single_api_callback(API.GET_CONTACT_POINTS, body_1_name, body_2_name)
        return self._parse_contact_points(body_1_name, body_2_name, points)

    def get_objects_intersected_with_rays(self, from_positions: List[List[float]],
                                          to_positions: List[List[float]]) -> List[MultiverseRayResult]:
        """
        Get the rays intersections with the objects from the from_positions to the to_positions.

        :param from_positions: The starting positions of the rays.
        :param to_positions: The ending positions of the rays.
        :return: The rays intersections with the objects as a list of MultiverseRayResult.
        """
        get_rays_response = self._get_rays(from_positions, to_positions)
        return self._parse_get_rays_response(get_rays_response)

    def _get_rays(self, from_positions: List[List[float]],
                  to_positions: List[List[float]]) -> List[str]:
        """
        Get the rays intersections with the objects from the from_positions to the to_positions.

        :param from_positions: The starting positions of the rays.
        :param to_positions: The ending positions of the rays.
        :return: The rays intersections with the objects as a dictionary.
        """
        from_positions = self.list_of_positions_to_string(from_positions)
        to_positions = self.list_of_positions_to_string(to_positions)
        return self._request_single_api_callback(API.GET_RAYS, from_positions, to_positions)

    @staticmethod
    def _parse_get_rays_response(response: List[str]) -> List[MultiverseRayResult]:
        """
        Parse the response of the get rays API.

        :param response: The response of the get rays API as a list of strings.
        :return: The rays as a list of lists of floats.
        """
        get_rays_results = []
        for ray_response in response:
            if ray_response == "None":
                get_rays_results.append(MultiverseRayResult("", -1))
            else:
                result = ray_response.split()
                result[1] = float(result[1])
                get_rays_results.append(MultiverseRayResult(*result))
        return get_rays_results

    @staticmethod
    def list_of_positions_to_string(positions: List[List[float]]) -> str:
        """
        Convert the list of positions to a string.

        :param positions: The list of positions.
        :return: The list of positions as a string.
        """
        return " ".join([f"{position[0]} {position[1]} {position[2]}" for position in positions])

    @staticmethod
    def _parse_constraint_effort(contact_effort: List[str]) -> Tuple[List[float], List[float]]:
        """
        Parse the contact effort of an object.

        :param contact_effort: The contact effort of the object as a list of strings.
        :return: The contact effort of the object as a list of floats.
        """
        contact_effort = contact_effort[0].split()
        contact_effort = list(map(float, contact_effort))
        forces, torques = contact_effort[:3], contact_effort[3:]
        return forces, torques

    @staticmethod
    def _parse_contact_points(body_1, body_2, contact_points: List[str]) -> List[MultiverseContactPoint]:
        """
        Parse the contact points of an object.

        :param body_1: The name of the first body.
        :param body_2: The name of the second body.
        :param contact_points: The contact points of the object as a list of strings.
        :return: The contact positions, and normal vectors as a list of MultiverseContactPoint.
        """
        contact_point_data = [list(map(float, contact_point.split())) for contact_point in contact_points]
        return [MultiverseContactPoint(body_1, body_2, point[:3], point[3:]) for point in contact_point_data]

    def get_contact_points(self, body_name: str) -> List[MultiverseContactPoint]:
        """
        Get the contact points of a body from the multiverse server.

        :param body_name: The name of the body.
        """
        contact_data = self._request_single_api_callback(API.GET_CONTACT_BODIES_AND_POINTS,
                                                         body_name, "with_children")
        contact_points = []
        for contact_point in contact_data:
            contact_point_data = list(contact_point.split())
            position_and_normal = list(map(float, contact_point_data[2:]))
            for i in range(0, len(position_and_normal), 6):
                contact_points.append(MultiverseContactPoint(contact_point_data[0],
                                                             contact_point_data[1],
                                                             position_and_normal[i:i + 3],
                                                             position_and_normal[i + 3:i + 6]))
        return contact_points

    def pause_simulation(self) -> None:
        """
        Pause the simulation.
        """
        self._request_single_api_callback(API.PAUSE)

    def unpause_simulation(self) -> None:
        """
        Unpause the simulation.
        """
        self._request_single_api_callback(API.UNPAUSE)

    def _request_single_api_callback(self, api_name: API, *params) -> List[str]:
        """
        Request a single API callback from the server.

        :param api_data: The API data to request the callback.
        :return: The API response as a list of strings.
        """
        response = self._request_apis_callbacks({api_name: list(params)})
        return response[api_name]

    def _request_apis_callbacks(self, api_data: Dict[API, List]) -> Dict[API, List[str]]:
        """
        Request the API callbacks from the server.

        :param api_data: The API data to add to the request metadata.
        :return: The API response as a list of strings.
        """
        self._reset_api_callback()
        for api_name, params in api_data.items():
            self._add_api_request(api_name.value, *params)
        self._send_api_request()
        if self.wait:
            sleep(self.API_REQUEST_WAIT_TIME.total_seconds() * self.simulation_wait_time_factor)
            self.wait = False
        responses = self._get_all_apis_responses()
        self.validate_apis_response(api_data, responses)
        return responses

    @staticmethod
    def validate_apis_response(api_data: Dict[API, List], responses: Dict[API, List[str]]):
        """
        Validate the responses from the multiverse server and raise error if an api request failed.

        :param api_data: The data of the api request which has the api name and the arguments.
        :param responses: The responses of the given api requests.
        :raises FailedAPIResponse: when one of the responses reports that the request failed.
        """
        for api_name, response in responses.items():
            for val in response:
                if 'failed' in val:
                    raise MultiverseFailedAPIResponse(response, api_name, api_data[api_name])

    def _get_all_apis_responses(self) -> Dict[API, List[str]]:
        """
        Get all the API responses from the server.

        :return: The API responses as a list of APIData.
        """
        list_of_api_responses = self.response_meta_data["api_callbacks_response"][self.simulation]
        return {API[api_name.upper()]: response for api_response in list_of_api_responses
                for api_name, response in api_response.items()}

    def _add_api_request(self, api_name: str, *params):
        """
        Add an API request to the request metadata.

        :param api_name: The name of the API.
        :param params: The parameters of the API.
        """
        self.request_meta_data["api_callbacks"][self.simulation].append({api_name: list(params)})

    def _send_api_request(self):
        """
        Send the API request to the server.
        """
        if "api_callbacks" not in self.request_meta_data:
            logging.error("No API request to send")
            raise ValueError
        self.send_and_receive_meta_data()
        self.request_meta_data.pop("api_callbacks")

    def _reset_api_callback(self):
        """
        Initialize the API callback in the request metadata.
        """
        self.request_meta_data["api_callbacks"] = {self.simulation: []}



from types import MethodType

import numpy as np

from ..datastructures.dataclasses import VirtualMobileBaseJoints
from ..pose_generator_and_validator import PoseGenerator
from ..ros import get_ros_package_path
from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager, CameraDescription
from ..datastructures.enums import Arms, Grasp, GripperState, TorsoState, GripperType, StaticJointState
from ..tf_transformations import quaternion_from_euler
from ..units import meter

filename = get_ros_package_path('pycram') + '/resources/robots/' + "iai_donbot" + '.urdf'

donbot_description = RobotDescription("iai_donbot", "base_link", "ur5_base_link", "arm_base_mounting_joint",
                                      filename, virtual_mobile_base_joints=VirtualMobileBaseJoints())

################################## Right Arm ##################################
left_arm = KinematicChainDescription("left_arm", "ur5_base_link", "ur5_wrist_3_link", donbot_description.urdf_object,
                                     arm_type=Arms.LEFT)

left_arm.add_static_joint_states("looking", {'ur5_shoulder_pan_joint': 0,
                                             'ur5_shoulder_lift_joint': -0.35,
                                             'ur5_elbow_joint': -2.15,
                                             'ur5_wrist_1_joint': -0.7,
                                             'ur5_wrist_2_joint': 1.57,
                                             'ur5_wrist_3_joint': -1.57})

left_arm.add_static_joint_states(StaticJointState.Park, {'ur5_shoulder_pan_joint': 3.23,
                                                         'ur5_shoulder_lift_joint': -1.51,
                                                         'ur5_elbow_joint': -1.57,
                                                         'ur5_wrist_1_joint': 0,
                                                         'ur5_wrist_2_joint': 1.57,
                                                         'ur5_wrist_3_joint': -1.65})

donbot_description.add_kinematic_chain_description(left_arm)

################################## Right Gripper ##################################

left_gripper = EndEffectorDescription("left_gripper", "gripper_base_link", "gripper_tool_frame",
                                      donbot_description.urdf_object)

left_gripper.add_static_joint_states(GripperState.OPEN, {'gripper_joint': 0.109,
                                                         "gripper_base_gripper_left_joint": -0.055})
left_gripper.add_static_joint_states(GripperState.CLOSE, {'gripper_joint': 0.0065,
                                                          "gripper_base_gripper_left_joint": -0.0027})
left_gripper.end_effector_type = GripperType.PARALLEL
left_gripper.opening_distance = 0.11 * meter  # 2x 55mm for WSG050
left_arm.end_effector = left_gripper

################################## Torso ##################################
torso = KinematicChainDescription("torso", "base_footprint", "ur5_base_link",
                                  donbot_description.urdf_object, include_fixed_joints=True)

# fixed joint, so all states set to 0
torso.add_static_joint_states(TorsoState.HIGH, {"arm_base_mounting_joint": 0})

torso.add_static_joint_states(TorsoState.MID, {"arm_base_mounting_joint": 0})

torso.add_static_joint_states(TorsoState.LOW, {"arm_base_mounting_joint": 0})

donbot_description.add_kinematic_chain_description(torso)

################################## Camera ##################################
camera = CameraDescription("camera_link", "camera_link", 0.75049, 0.5, 1.2)
donbot_description.add_camera_description(camera)

################################## Neck ##################################
donbot_description.add_kinematic_chain("neck", "ur5_base_link", "ur5_base_link")

################################# Grasps ##################################
left_gripper.update_all_grasp_orientations([0.707, -0.707, 0.707, -0.707])


################################### Custom Orientation Generator ##############
def donbot_orientation_generator(position, origin):
    angle = np.arctan2(position[1] - origin.position.y, position[0] - origin.position.x) - np.pi / 2

    quaternion = list(quaternion_from_euler(0, 0, angle, axes="sxyz"))
    return quaternion


################################ Load Function #################################
def load(description):
    RobotDescription.current_robot_description = description
    PoseGenerator.override_orientation_generator = donbot_orientation_generator


def unload(description):
    RobotDescription.current_robot_description = None
    PoseGenerator.override_orientation_generator = None


donbot_description.load = MethodType(load, donbot_description)
donbot_description.unload = MethodType(unload, donbot_description)

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(donbot_description)



from owlready2 import Imp

from .dependencies import *


class World(Base):
    """
    The world could be the belief state of an agent about the world.
    """


class Floor(Base):
    """
    The floor of the world.
    """


class Milk(Base):
    """
    A milk carton.
    """


class Robot(Base):
    """
    The robot.
    """


class Cereal(Base):
    """
    A cereal box.
    """


class Kitchen(Base):
    """
    A kitchen.
    """


class Food(Base):
    ...


class Fruit(Food):
    ...


class Apple(Fruit):
    ...


class Environment(Base):
    ...


class Apartment(Environment):
    ...


class Cup(Base):
    ...


class Spoon(Base):
    ...


class Bowl(Base):
    ...


class PreferredGraspAlignment(Base):
    ...


class XAxis(Base):
    value = (1, 0, 0)


class YAxis(Base):
    value = (0, 1, 0)


class NoAlignment(Base):
    value = (0, 0, 0)


class Truthy(Base):
    value = True


class Falsy(Base):
    value = False


class Cabinet(Base):
    ...


class Washer(Base):
    ...


class Drawer(Base):
    ...


class Refrigerator(Base):
    ...


class Sink(Base):
    ...


class Door(Base):
    ...


class Cutting(Base):
    ...


class Pouring(Base):
    ...


class Handle(Base):
    ...


class Link(Base):
    ...


class PhysicalObject(Base):
    ...


class PouringTool(PhysicalObject):
    """
    The Tool that is used for pouring, can be cup, bottle, etc.
    """


class CuttingTool(PhysicalObject):
    """
    The tool used for cutting — such as a knife, bread-knife, or similar — depends on the specific task and object being processed.
    """


class MixingTool(PhysicalObject):
    """
    The tool used for mixing — such as a Spoon, Whisk, or similar — depends on the specific task and object being processed.
    """


class Agent(Base):
    ...


class Human(Base):
    ...


class Room(Base):
    ...


class Location(Base):
    ...


class Container(Base):
    ...


class Joint(Base):
    ...


class ContinuousJoint(Base):
    """
    A continuous hinge joint that rotates around an axis and has no upper and lower limits.
    """


class HingeJoint(Base):
    """
    A joint that rotates along an axis.
    """


class FixedJoint(Base):
    """
    A joint that cannot move, designed to fixiate links.
    """


class MovableJoint(Base):
    """
    A joint where the two connected links can move relative to each other in some dimension.
    """


class FloatingJoint(Base):
    """
    A joint that allows motion for all 6 degrees of freedom.
    """


class PlanarJoint(Base):
    """
    A joint that allows motion in a plane perpendicular to an axis.
    """


class PrismaticJoint(Base):
    """
    A sliding joint that slides along an axis, and has a limited range specified by the upper and lower limits.
    """


class RevoluteJoint(Base):
    """
    A hinge joint that rotates along an axis and has a limited range specified by the upper and lower limits.
    """


class DesignedFurniture(Base):
    """
    An object used to make a room or building suitable for living or working.
    """


class Surface(Base):
    """
    The outside part or uppermost layer of something.
    """


class PhysicalTask(Base):
    """
    A task in which a PhysicalAgent affects some physical object.
    """


class Action(Base):
    """
    An Event with at least one Agent that isParticipantIn it, and that executes a Task that typically isDefinedIn a Plan, Workflow, Project, etc.
    """


class Event(Base):
    """
    Any physical, social, or mental process, event, or state.

    More theoretically, events can be classified in different ways, possibly based on 'aspect' (e.g. stative, continuous, accomplishement, achievement, etc.), on 'agentivity' (e.g. intentional, natural, etc.), or on 'typical participants' (e.g. human, physical, abstract, food, etc.).
    Here no special direction is taken, and the following explains why: events are related to observable situations, and they can have different views at a same time.
    If a position has to be suggested here anyway, the participant-based classification of events seems the most stable and appropriate for many modelling problems.

    (1) Alternative aspectual views

    Consider a same event 'rock erosion in the Sinni valley': it can be conceptualized as an accomplishment (what has brought a certain state to occur), as an achievement (the state resulting from a previous accomplishment), as a punctual event (if we collapse the time interval of the erosion into a time point), or as a transition (something that has changed from a state to a different one).
    In the erosion case, we could therefore have good motivations to shift from one aspect to another: a) causation focus, b) effectual focus, c) historical condensation, d) transition (causality).

    The different views refer to the same event, but are still different: how to live with this seeming paradox?
    A typical solution e.g. in linguistics (cf. Levin's aspectual classes) and in DOLCE Full (cf. WonderWeb D18 axiomatization) is to classify events based on aspectual differences. But this solution would create different identities for a same event, where the difference is only based on the modeller's attitude.
    An alternative solution is suggested here, and exploits the notion of (observable) Situation; a Situation is a view, consistent with a Description, which can be observed of a set of entities. It can also be seen as a 'relational context' created by an observer on the basis of a 'frame'. Therefore, a Situation allows to create a context where each particular view can have a proper identity, while the Event preserves its own identity.
    For example, ErosionAsAccomplishment is a Situation where rock erosion is observed as a process leading to a certain achievement: the conditions (roles, parameters) that suggest such view are stated in a Description, which acts as a 'theory of accomplishments'. Similarly, ErosionAsTransition is a Situation where rock erosion is observed as an event that has changed a state to another: the conditions for such interpretation are stated in a different Description, which acts as a 'theory of state transitions'.
    Consider that in no case the actual event is changed or enriched in parts by the aspectual view.

    (2) Alternative intentionality views

    Similarly to aspectual views, several intentionality views can be provided for a same Event. For example, one can investigate if an avalanche has been caused by immediate natural forces, or if there is any hint of an intentional effort to activate those natural forces.
    Also in this case, the Event as such has not different identities, while the causal analysis generates situations with different identities, according to what Description is taken for interpreting the Event.
    On the other hand, if the possible actions of an Agent causing the starting of an avalanche are taken as parts of the Event, then this makes its identity change, because we are adding a part to it.
    Therefore, if intentionality is a criterion to classify events or not, this depends on if an ontology designer wants to consider causality as a relevant dimension for events' identity.

    (3) Alternative participant views

    A slightly different case is when we consider the basic participants to an Event. In this case, the identity of the Event is affected by the participating objects, because it depends on them.
    For example, if snow, mountain slopes, wind, waves, etc. are considered as an avalanche basic participants, or if we also want to add water, human agents, etc., that makes the identity of an avalanche change.
    Anyway, this approach to event classification is based on the designer's choices, and more accurately mirrors lexical or commonsense classifications (see. e.g. WordNet 'supersenses' for verb synsets).

    Ultimately, this discussion has no end, because realists will keep defending the idea that events in reality are not changed by the way we describe them, while constructivists will keep defending the idea that, whatever 'true reality' is about, it can't be modelled without the theoretical burden of how we observe and describe it.
    Both positions are in principle valid, but, if taken too radically, they focus on issues that are only partly relevant to the aim of computational ontologies, which assist domain experts in representing a certain portion of reality according to their own assumptions and requirements.

    For this reason, in this ontology version of DOLCE, both events and situations are allowed, together with descriptions (the reason for the inclusion of the D&S framewrok in DOLCE), in order to encode the modelling needs, independently from the position (if any) chosen by the model designer.
    """


class Entity(Base):
    """
    Anything: real, possible, or imaginary, which some modeller wants to talk about for some purpose.
    """


class Task(Base):
    """
    An EventType that classifies an Action to be executed.
    For example, reaching a destination is a task that can be executed by performing certain actions, e.g. driving a car, buying a train ticket, etc.
    The actions to execute a task can also be organized according to a Plan that is not the same as the one that defines the task (if any).
    For example, reaching a destination could be defined by a plan to get on holidays, while the plan to execute the task can consist of putting some travels into a sequence.
    """


class RootLink(Base):
    ...


class Supporter(Base):
    """
    A physical object that supports another object.
    """
    ...


class SupportedObject(Base):
    """
    A physical object that is supported by another object.
    """
    ...



import actionlib

from actionlib import SimpleActionClient

def create_action_client(topic_name: str, action_message) -> SimpleActionClient:
    return actionlib.SimpleActionClient(topic_name, action_message)




# TODO fix views for new imperative orm structure
# from sqlalchemy.orm import declarative_base, Mapped, column_property
# from typing_extensions import Union
# import sqlalchemy.orm
# from sqlalchemy import table, inspect, event, select, engine, MetaData, Select, TableClause, ExecutableDDLElement
# from sqlalchemy.ext.compiler import compiles
# from .object_designator import Object
#
#
# class CreateView(ExecutableDDLElement):
#     """
#     Class that is used to create a view. Every instance will be compiled into a ORM CREATE VIEW statement.
#     """
#
#     def __init__(self, name: str, selectable: Select):
#         self.name = name
#         self.selectable = selectable
#
#
# class DropView(ExecutableDDLElement):
#     """
#     Class that is used to drop a view. Every instance will be compiled into a ORM DROP VIEW statement.
#     """
#
#     def __init__(self, name: str):
#         self.name = name
#
#
# @compiles(CreateView)
# def _create_view(element: CreateView, compiler, **kw) -> str:
#     """
#     Compiles a CreateView instance into a SQL CREATE VIEW statement.
#     :param element: CreateView instance
#     :param compiler: compiler
#     :param kw: keyword arguments
#     :return: SQL CREATE VIEW statement
#     """
#
#     return "CREATE VIEW %s AS %s" % (
#         element.name,
#         compiler.sql_compiler.process(element.selectable, literal_binds=True),
#     )
#
#
# @compiles(DropView)
# def _drop_view(element: DropView, compiler, **kw) -> str:
#     """
#     Compiles a DropView instance into a SQL DROP VIEW statement.
#     :param element: DropView instance
#     :param compiler: compiler
#     :param kw: keyword arguments
#     :return: SQL DROP VIEW statement
#     """
#     return "DROP VIEW %s" % element.name
#
#
# def view_exists(ddl: Union[CreateView, DropView], target, connection: engine, **kw) -> bool:
#     """
#     Check if a view exists.
#     :param ddl: ddl instance
#     :param target: target object
#     :param connection: connection
#     :param kw: keyword arguments
#     :return: True if the view exists, False otherwise
#     """
#
#     return ddl.name in inspect(connection).get_view_names()
#
#
# def view_doesnt_exist(ddl: Union[CreateView, DropView], target, connection: engine, **kw) -> bool:
#     """
#     Check if a view does not exist.
#     :param ddl: ddl instance
#     :param target: target object
#     :param connection: connection
#     :param kw: keyword arguments
#     :return: True if the view does not exist, False otherwise
#     """
#
#     return not view_exists(ddl, target, connection, **kw)
#
#
# def view(name: str, metadata: MetaData, selectable: Select) -> TableClause:
#     """
#     Function used to control view creation and deletion. It will listen to the after_create and before_drop events
#     of the metadata object in order to either create or drop the view. The view needs to have a column id.
#     """
#     view = table(name)
#
#     view._columns._populate_separate_keys(
#         col._make_proxy(view) for col in selectable.selected_columns
#     )
#
#     event.listen(metadata, "after_create", CreateView(name, selectable).execute_if(callable_=view_doesnt_exist))
#     event.listen(metadata, "before_drop", DropView(name).execute_if(callable_=view_exists))
#
#     return view
#
#
# base = declarative_base(metadata=Base.metadata)
#
#
# class PickUpWithContextView(base):
#     """
#     View for pickup performables with context.
#     """
#
#     __robot_position: Position = sqlalchemy.orm.aliased(Position, flat=True)
#     """
#     3D Vector of robot position
#     """
#
#     __robot_pose: Pose = sqlalchemy.orm.aliased(Pose, flat=True)
#     """
#     Complete robot pose
#     """
#
#     __object_position: Position = sqlalchemy.orm.aliased(Position, flat=True)
#     """
#     3D Vector for object position
#     """
#
#     __table__ = view("PickUpWithContextView", Base.metadata,
#                      (select(PickUpAction.id, PickUpAction.arm, GraspDescription.approach_direction,
#                              GraspDescription.vertical_alignment, GraspDescription.rotate_gripper,
#                              RobotState.torso_height, (__robot_position.x-__object_position.x).label("relative_x"),
#                              (__robot_position.y-__object_position.y).label("relative_y"), Quaternion.x, Quaternion.y,
#                              Quaternion.z, Quaternion.w, Object.obj_type, TaskTreeNode.status)
#                       .join(TaskTreeNode.action.of_type(PickUpAction))
#                       .join(PickUpAction.robot_state)
#                       .join(__robot_pose, RobotState.pose)
#                       .join(__robot_position, __robot_pose.position)
#                       .join(Pose.orientation)
#                       .join(PickUpAction.object)
#                       .join(Object.pose)
#                       .join(__object_position, Pose.position)))
#
#     id: Mapped[int] = __table__.c.id
#     arm: Mapped[str] = __table__.c.arm
#     torso_height: Mapped[float] = __table__.c.torso_height
#     relative_x: Mapped[float] = column_property(__table__.c.relative_x)
#     relative_y: Mapped[float] = column_property(__table__.c.relative_y)
#     quaternion_x: Mapped[float] = __table__.c.x
#     quaternion_y: Mapped[float] = __table__.c.y
#     quaternion_z: Mapped[float] = __table__.c.z
#     quaternion_w: Mapped[float] = __table__.c.w
#     obj_type: Mapped[str] = __table__.c.obj_type
#     status: Mapped[str] = __table__.c.status
#
#
# class PlaceWithContextView(base):
#     """
#     View for pickup performables with context.
#     """
#
#     __robot_position: Position = sqlalchemy.orm.aliased(Position, flat=True)
#     """
#     3D Vector of robot position
#     """
#
#     __robot_pose: Pose = sqlalchemy.orm.aliased(Pose, flat=True)
#     """
#     Complete robot pose
#     """
#
#     __object_position: Position = sqlalchemy.orm.aliased(Position, flat=True)
#     """
#     3D Vector for object position
#     """
#
#     __table__ = view("PlaceWithContextView", Base.metadata,
#                      (select(PlaceAction.id, PlaceAction.arm, RobotState.torso_height,
#                              (__robot_position.x-__object_position.x).label("relative_x"),
#                              (__robot_position.y-__object_position.y).label("relative_y"), Quaternion.x, Quaternion.y,
#                              Quaternion.z, Quaternion.w, Object.obj_type, TaskTreeNode.status)
#                       .join(TaskTreeNode.action.of_type(PlaceAction))
#                       .join(PlaceAction.robot_state)
#                       .join(__robot_pose, RobotState.pose)
#                       .join(__robot_position, __robot_pose.position)
#                       .join(Pose.orientation)
#                       .join(PlaceAction.object)
#                       .join(Object.pose)
#                       .join(__object_position, Pose.position)))
#
#     id: Mapped[int] = __table__.c.id
#     arm: Mapped[str] = __table__.c.arm
#     torso_height: Mapped[float] = __table__.c.torso_height
#     relative_x: Mapped[float] = column_property(__table__.c.relative_x)
#     relative_y: Mapped[float] = column_property(__table__.c.relative_y)
#     quaternion_x: Mapped[float] = __table__.c.x
#     quaternion_y: Mapped[float] = __table__.c.y
#     quaternion_z: Mapped[float] = __table__.c.z
#     quaternion_w: Mapped[float] = __table__.c.w
#     obj_type: Mapped[str] = __table__.c.obj_type
#     status: Mapped[str] = __table__.c.status




from __future__ import print_function

import sys
import json
from typing import Callable
from ..ros import logwarn, loginfo, get_node_names, wait_for_service, get_service_proxy, ServiceException
try:
    from iai_apartment_kitchen_msgs.srv import Authenticateuser
except ModuleNotFoundError as e:
    pass


def init_kitchen_interface(func: Callable) -> Callable:
    """
    Tries to import the messages and initialize interface.
    """
    def wrapper(*args, **kwargs):
        if "iai_apartment_kitchen_msgs" not in sys.modules:
            logwarn("Could not initialize the Apartment kitchen interface since the iai_apartment_kitchen_msgs are not imported")
            return

        if "/blum_control_service_server" in get_node_names():
            loginfo("Successfully initialized Apartment kitchen interface")
        else:
            logwarn("Apartment kitchen is not running, could not initialize Apartment kitchen interface")
            return
        authenticate_user()

        return func(*args, **kwargs)
    return wrapper


def authenticate_user():
    """
    Authenticates a user at the bridge
    """
    loginfo('Waiting for authentication')
    wait_for_service('/blum_kitchen_server')

    try:
        kitchen_service = get_service_proxy('/blum_kitchen_server', Authenticateuser)
        resp1 = kitchen_service('authenticate_chk', 'pycram')
        if resp1.resp == 'FALSE':
            loginfo('Please press the cyan button on the kitchen gateway')
            resp1 = kitchen_service('authenticate', 'pycram')
    except ServiceException as e:
        print(f"Service call failed: {e}")

@init_kitchen_interface
def call_kitchen_service(command, argument):
    loginfo('Waiting for kitchen service')
    wait_for_service('/blum_kitchen_server')

    try:
        kitchen_service = get_service_proxy('/blum_kitchen_server', Authenticateuser)
        resp1 = kitchen_service(command, argument)
        return resp1.resp
    except ServiceException as e:
        print(f"Service call failed: {e}")


@init_kitchen_interface
def open_cabinet(kitchen_element):
    """
    Opens a cabinet of the apartment kitchen

    :param kitchen_element: The cabinet which should be opened
    """
    call_kitchen_service('open', kitchen_element)


@init_kitchen_interface
def close_cabinet(kitchen_element):
    """
    Closes a cabinet of the apartment kitchen, currently this is only possible for the "oberschrank"

    :param kitchen_element: Cabinet which should be closed
    """
    call_kitchen_service('close', kitchen_element)

@init_kitchen_interface
def modules_show():
    """
    Shows all possible cabinet modules that can be called
    :return:
    """
    return call_kitchen_service('modules_show', ' ')


@init_kitchen_interface
def kitchen_state() -> Dict[str, Dict[str, str]]:
    """
    Returns the state of the whole kitchen as a dictionary. The mapping is "name": {"state", "type"}

    :return: A dict describing the kitchen state
    """
    state = call_kitchen_service("kitchen_state_json"," ")
    state_dict = json.loads(state.replace("'", '"'))
    return state_dict


from ..datastructures.dataclasses import VirtualMobileBaseJoints
from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager, CameraDescription
from ..datastructures.enums import Arms, Grasp, GripperState, GripperType, TorsoState, StaticJointState
from ..ros import  get_ros_package_path
from ..units import meter

from ..helper import get_robot_description_path

filename = get_ros_package_path('pycram') + '/resources/robots/' + "pr2" + '.urdf'

mjcf_filename = get_robot_description_path("", "pr2")

pr2_description = RobotDescription("pr2", "base_link", "torso_lift_link",
                                   "torso_lift_joint", filename,
                                   virtual_mobile_base_joints=VirtualMobileBaseJoints(),
                                   mjcf_path=mjcf_filename,
                                   ignore_joints=['torso_lift_motor_screw_joint', 'r_gripper_motor_slider_joint',
                                                  'r_gripper_motor_screw_joint', 'r_gripper_joint',
                                                  'l_gripper_motor_slider_joint', 'l_gripper_motor_screw_joint',
                                                  'l_gripper_joint'])

################################## Left Arm ##################################
left_arm = KinematicChainDescription("left", "torso_lift_link", "l_wrist_roll_link",
                                     pr2_description.urdf_object, arm_type=Arms.LEFT)
left_arm.add_static_joint_states(StaticJointState.Park, {'l_shoulder_pan_joint': 1.712,
                                          'l_shoulder_lift_joint': -0.264,
                                          'l_upper_arm_roll_joint': 1.38,
                                          'l_elbow_flex_joint': -2.12,
                                          'l_forearm_roll_joint': 16.996,
                                          'l_wrist_flex_joint': -0.073,
                                          'l_wrist_roll_joint': 0.0})
pr2_description.add_kinematic_chain_description(left_arm)

################################## Left Gripper ##################################
left_gripper = EndEffectorDescription("left_gripper", "l_gripper_palm_link", "l_gripper_tool_frame",
                                      pr2_description.urdf_object,
                                      fingers_link_names=['l_gripper_l_finger_tip_link', 'l_gripper_r_finger_tip_link'])
left_gripper.add_static_joint_states(GripperState.OPEN, {'l_gripper_l_finger_joint': 0.548,
                                                         'l_gripper_r_finger_joint': 0.548})
left_gripper.add_static_joint_states(GripperState.CLOSE, {'l_gripper_l_finger_joint': 0.0,
                                                          'l_gripper_r_finger_joint': 0.0})
left_gripper.end_effector_type = GripperType.PARALLEL
left_gripper.opening_distance = 0.086 * meter
left_arm.end_effector = left_gripper

################################## Right Arm ##################################
right_arm = KinematicChainDescription("right", "torso_lift_link", "r_wrist_roll_link",
                                      pr2_description.urdf_object, arm_type=Arms.RIGHT)
right_arm.add_static_joint_states(StaticJointState.Park, {'r_shoulder_pan_joint': -1.712,
                                           'r_shoulder_lift_joint': -0.256,
                                           'r_upper_arm_roll_joint': -1.463,
                                           'r_elbow_flex_joint': -2.12,
                                           'r_forearm_roll_joint': 1.766,
                                           'r_wrist_flex_joint': -0.07,
                                           'r_wrist_roll_joint': 0.051})
pr2_description.add_kinematic_chain_description(right_arm)

################################## Right Gripper ##################################
right_gripper = EndEffectorDescription("right_gripper", "r_gripper_palm_link", "r_gripper_tool_frame",
                                       pr2_description.urdf_object,
                                       fingers_link_names=['r_gripper_l_finger_tip_link', 'r_gripper_r_finger_tip_link'])
right_gripper.add_static_joint_states(GripperState.OPEN, {'r_gripper_l_finger_joint': 0.548,
                                                          'r_gripper_r_finger_joint': 0.548})
right_gripper.add_static_joint_states(GripperState.CLOSE, {'r_gripper_l_finger_joint': 0.0,
                                                           'r_gripper_r_finger_joint': 0.0})
right_gripper.end_effector_type = GripperType.PARALLEL
right_gripper.opening_distance = 0.086 * meter
right_arm.end_effector = right_gripper

################################## Camera ##################################
camera = CameraDescription("kinect_camera", "wide_stereo_optical_frame", 1.27,
                           1.60, 0.99483, 0.75049)
pr2_description.add_camera_description(camera)

################################## Torso ##################################
torso = KinematicChainDescription("torso", "base_link", "torso_lift_link",
                                  pr2_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"torso_lift_joint": 0.3})

torso.add_static_joint_states(TorsoState.MID, {"torso_lift_joint": 0.15})

torso.add_static_joint_states(TorsoState.LOW, {"torso_lift_joint": 0})

pr2_description.add_kinematic_chain_description(torso)

################################## Neck ##################################
pr2_description.set_neck(yaw_joint="head_pan_joint", pitch_joint="head_tilt_joint")

################################# Grasps ##################################
left_gripper.update_all_grasp_orientations([0, 0, 0, 1])
left_gripper.set_approach_axis([1, 0, 0])
right_gripper.update_all_grasp_orientations([0, 0, 0, 1])
right_gripper.set_approach_axis([1, 0, 0])

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(pr2_description)



from __future__ import annotations

import inspect
import time
from dataclasses import field, dataclass
from datetime import datetime

import networkx as nx
from typing_extensions import Optional, Callable, Any, Dict, List, Iterable, TYPE_CHECKING, Type, Tuple, Iterator
from random_events.product_algebra import Event

from typing_extensions import Optional, Callable, Any, Dict, List, Iterable, TYPE_CHECKING, Type, Tuple, Iterator

from .datastructures.enums import TaskStatus
from .external_interfaces import giskard
from .failures import PlanFailure
from .has_parameters import leaf_types
from .ros import loginfo

if TYPE_CHECKING:
    from .designator import BaseMotion, ActionDescription


class Plan(nx.DiGraph):
    """
    Represents a plan structure, typically a tree, which can be changed at any point in time. Performing the plan will
    traverse the plan structure in depth first order and perform each PlanNode
    """
    current_plan: Plan = None

    on_start_callback: Dict[Optional[Type[ActionDescription]], List[Callable]] = {}
    on_end_callback: Dict[Optional[Type[ActionDescription]], List[Callable]] = {}

    def __init__(self, root: PlanNode):
        super().__init__()
        self.root: PlanNode = root
        self.add_node(self.root)
        self.current_node: PlanNode = self.root
        self.on_start_callback = {}
        self.on_end_callback = {}

    def mount(self, other: Plan, mount_node: PlanNode = None):
        """
        Mounts another plan to this plan. The other plan will be added as a child of the mount_node.

        :param other: The plan to be mounted
        :param mount_node: A node of this plan to which the other plan will be mounted. If None, the root of this plan will be used.
        """
        mount_node = mount_node or self.root
        self.add_nodes_from(other.nodes)
        self.add_edges_from(other.edges)
        self.add_edge(mount_node, other.root)
        for node in self.nodes:
            node.plan = self

    def merge_nodes(self, node1: PlanNode, node2: PlanNode):
        """
        Merges two nodes into one. The node2 will be removed and all its children will be added to node1.

        :param node1: Node which will remain in the plan
        :param node2: Node which will be removed from the plan
        """
        for node in node2.children:
            self.add_edge(node1, node)
        self.remove_node(node2)

    def add_node(self, node_for_adding: PlanNode, **attr):
        """
        Adds a node to the plan. The node will not be connected to any other node of the plan.

        :param node_for_adding: Node to be added
        :param attr: Additional attributes to be added to the node
        """
        super().add_node(node_for_adding, **attr)
        node_for_adding.plan = self

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        """
        Adds an edge to the plan. If one or both nodes are not in the plan, they will be added to the plan.

        :param u_of_edge: Origin node of the edge
        :param v_of_edge: Target node of the edge
        :param attr: Additional attributes to be added to the edge
        """
        super().add_edge(u_of_edge, v_of_edge, **attr)
        u_of_edge.plan = self
        v_of_edge.plan = self

    def add_edges_from(self, ebunch_to_add: Iterable[Tuple[PlanNode, PlanNode]], **attr):
        """
        Adds edges to the plan from an iterable of tuples. If one or both nodes are not in the plan, they will be added to the plan.

        :param ebunch_to_add: Iterable of tuples of nodes to be added
        :param attr: Additional attributes to be added to the edges
        """
        super().add_edges_from(ebunch_to_add, **attr)
        for u_edge, v_edge in ebunch_to_add:
            u_edge.plan = self
            v_edge.plan = self

    def add_nodes_from(self, nodes_for_adding: Iterable[PlanNode], **attr):
        """
        Adds nodes from an Iterable of nodes.

        :param nodes_for_adding: The iterable of nodes
        :param attr: Addotional attributes to be added
        """
        super().add_nodes_from(nodes_for_adding, **attr)
        for node in nodes_for_adding:
            node.plan = self

    def insert_below(self, insert_node: PlanNode, insert_below: PlanNode):
        """
        Inserts a node below the given node.

        :param insert_node: The node to be inserted
        :param insert_below: A node of the plan below which the given node should be added
        """
        self.add_edge(insert_below, insert_node)

    def perform(self) -> Any:
        """
        Performs the root node of this plan.

        :return: The return value of the root node
        """
        previous_plan = Plan.current_plan
        Plan.current_plan = self
        result = self.root.perform()
        Plan.current_plan = previous_plan
        return result

    def resolve(self):
        """
        Resolves the root node of this plan if it is a DesignatorNode

        :return: The resolved designator
        """
        if isinstance(self.root, DesignatorNode):
            return self.root.designator_ref.resolve()

    def flattened_parameters(self):
        """
        The atomic parameter of this plan, as dict with paths as keys and the atomic type as value

        :return: A dict of the atomic types
        """
        result = {}
        for node in self.nodes:
            if isinstance(node, DesignatorNode):
                result.update(node.flattened_parameters())
        return result

    def re_perform(self):
        for child in self.root.recursive_children:
            if child.is_leaf:
                child.perform()

    @property
    def actions(self) -> List[ActionNode]:
        return list(filter(None, [node if type(node) is ActionNode else None for node in self.nodes]))

    def plot(self):
        """
        Plots the plan using matplotlib and networkx. The plan is plotted as a tree with the root node at the bottom and
        the children nodes above.
        """
        import matplotlib.pyplot as plt

        # Create a new figure
        plt.figure(figsize=(12, 8))

        # Use spring layout for node positioning
        pos = nx.drawing.bfs_layout(self, start=self.root, align='horizontal')

        # Draw nodes (bodies)
        nx.draw_networkx_nodes(self, pos,
                               node_color='lightblue',
                               node_size=2000)

        # Draw edges (connections)
        edges = self.edges(data=True)
        nx.draw_networkx_edges(self, pos,
                               edge_color='gray',
                               arrows=True,
                               arrowsize=20)

        # Add link names as labels
        labels = {node: str(node) for node in self.nodes()}
        nx.draw_networkx_labels(self, pos, labels)

        plt.title("Plan Structure")
        plt.axis('off')  # Hide axes
        plt.show()

    @classmethod
    def add_on_start_callback(cls, callback: Callable[[PlanNode], None],
                              action_type: Optional[Type[ActionDescription], Type[PlanNode]] = None):
        """
        Adds a callback to be called when an action of the given type is started.

        :param callback: The callback to be called
        :param action_type: The type of the action, if None, the callback will be called for all actions
        """
        if not cls.on_start_callback:
            cls.on_start_callback = {}
        if action_type not in cls.on_start_callback:
            cls.on_start_callback[action_type] = []
        cls.on_start_callback[action_type].append(callback)

    @classmethod
    def add_on_end_callback(cls, callback: Callable[[PlanNode], None],
                            action_type: Optional[Type[ActionDescription], Type[PlanNode]] = None):
        """
        Adds a callback to be called when an action of the given type is ended.

        :param callback: The callback to be called
        :param action_type: The type of the action
        """
        if not cls.on_end_callback:
            cls.on_end_callback = {}
        if action_type not in cls.on_end_callback:
            cls.on_end_callback[action_type] = []
        cls.on_end_callback[action_type].append(callback)

    @classmethod
    def remove_on_start_callback(cls, callback: Callable[[PlanNode], None],
                                 action_type: Optional[Type[ActionDescription], Type[PlanNode]] = None):
        """
        Removes a callback to be called when an action of the given type is started.

        :param callback: The callback to be removed
        :param action_type: The type of the action
        """
        if cls.on_start_callback and action_type in cls.on_start_callback:
            cls.on_start_callback[action_type].remove(callback)

    @classmethod
    def remove_on_end_callback(cls, callback: Callable[[PlanNode], None],
                               action_type: Optional[Type[ActionDescription], Type[PlanNode]] = None):
        """
        Removes a callback to be called when an action of the given type is ended.

        :param callback: The callback to be removed
        :param action_type: The type of the action
        """
        if cls.on_end_callback and action_type in cls.on_end_callback:
            cls.on_end_callback[action_type].remove(callback)

    def _create_pure_networkx_graph(self, attributes: List[str]) -> nx.DiGraph[int]:
        """
        Creates a pure networkx graph of this plan and adds the given attributes of nodes as networkx Node attributes.

        :param attributes: A list of attributes from the nodes which should be contained in the returned graph
        :return: A NetworkX graph from hash values of the PlanNodes

        """
        hash_nodes = {hash(node): node for node in self.nodes}
        edges = [(hash(source), hash(target)) for source, target in self.edges]
        graph = nx.DiGraph()
        graph.add_nodes_from(hash_nodes.keys())
        graph.add_edges_from(edges)
        node_colors = {TaskStatus.CREATED: "lightgrey", TaskStatus.RUNNING: "lightblue", TaskStatus.SUCCEEDED: "lightgreen",
                       TaskStatus.FAILED: "lightcoral", TaskStatus.INTERRUPTED: "lightpink", TaskStatus.SLEEPING: "lightyellow"}


        for v in graph:
            for attr in attributes:
                graph.nodes[v][attr] = str(getattr(hash_nodes[v], attr))
                graph.nodes[v]["node_type"] = hash_nodes[v].__class__.__name__
                graph.nodes[v]["node_color"] = node_colors[hash_nodes[v].status]
        return graph

    def plot_bokeh(self, attributes: List[str] = None):
        """
        Plots the plan using bokeh and networkx. The plan is plotted as a tree with the root node at the bottom and
        PlanNode.action attributes as labels.
        The plot features a hover tool showing the attributes of the nodes when the mouse is over them. Shown attributes
        can be configured using the attributes parameter. The attributes have to be a subset of the PlanNode attributes.

        :param attributes: A list of attributes from the nodes which should be shown in the hover tool.
        """
        attributes = attributes or ["status", "start_time"]
        from bokeh.plotting import figure, from_networkx, show
        from bokeh.models import (HoverTool, NodesAndLinkedEdges)

        p = figure(x_range=(-2, 2), y_range=(-2, 2),
                   width=1700, height=950,
                   x_axis_location=None, y_axis_location=None, toolbar_location="below",
                   title="Plan Visualization", background_fill_color="#efefef", )
        node_hover_tool = HoverTool(tooltips= [("node_type", "@node_type")] + [(attr, "@" + attr) for attr in attributes])
        p.add_tools(node_hover_tool)

        p.grid.grid_line_color = None
        p.add_layout(self._create_labels())

        graph = from_networkx(self._create_pure_networkx_graph(attributes), nx.drawing.bfs_layout,
                              start=hash(self.root), align='horizontal')
        graph.selection_policy = NodesAndLinkedEdges()
        graph.inspection_policy = NodesAndLinkedEdges()

        graph.node_renderer.glyph.update(size=20, fill_color="node_color")

        p.renderers.append(graph)

        show(p, new="same")

    def _create_labels(self):
        """
        Creates a label set for the plan visualization. Labels are the PlanNode.action attribute.

        :return: A LabelSet object which can be added to a bokeh plot.
        """
        from bokeh.models import ColumnDataSource, LabelSet
        hash_nodes = {hash(node): node for node in self.nodes}
        layout=nx.drawing.bfs_layout(self._create_pure_networkx_graph([]), start=hash(self.root), align='horizontal')
        x = [pose[0] for pose in layout.values()]
        y = [pose[1] for pose in layout.values()]
        name = [str(hash_nodes[node].action.__name__) for node in layout.keys()]
        label_dict = {'x': x, 'y': y, 'names': name}

        data_source = ColumnDataSource(data=label_dict)
        labels= LabelSet(x='x', y='y', text='names',
                          x_offset=-55, y_offset=10, source=data_source)
        return labels


def managed_node(func: Callable) -> Callable:
    """
    Decorator which manages the state of a node, including the start and end time, status and reason of failure as well
    as the setting of the current node in the plan.

    :param func: Reference to the perform function of the node
    :return: The wrapped perform function
    """

    def wrapper(node: DesignatorNode) -> Any:
        def wait(node):
            continue_execution = False
            while not continue_execution:
                all_parents_status = [parent.status for parent in node.all_parents] + [node.status]
                if TaskStatus.SLEEPING not in all_parents_status:
                    continue_execution = True
                time.sleep(0.1)

        all_parents_status = [parent.status for parent in node.all_parents] + [node.status]
        if TaskStatus.INTERRUPTED in all_parents_status:
            return
        elif TaskStatus.SLEEPING in all_parents_status:
            wait(node)

        node.status = TaskStatus.RUNNING
        node.start_time = datetime.now()
        on_start_callbacks = (Plan.on_start_callback.get(node.action, []) +
                              Plan.on_start_callback.get(None, []) + Plan.on_start_callback.get(node.__class__, []))
        on_end_callbacks = (Plan.on_end_callback.get(node.action, []) +
                            Plan.on_end_callback.get(None, []) + Plan.on_end_callback.get(node.__class__, []))
        for call_back in on_start_callbacks:
            call_back(node)
        result = None
        try:
            node.plan.current_node = node
            result = func(node)
            node.status = TaskStatus.SUCCEEDED
            node.result = result
        except PlanFailure as e:
            node.status = TaskStatus.FAILED
            node.reason = e
            raise e
        finally:
            node.end_time = datetime.now()
            node.plan.current_node = node.parent
            for call_back in on_end_callbacks:
                call_back(node)
        return result

    return wrapper


@dataclass
class PlanNode:
    status: TaskStatus = TaskStatus.CREATED
    """
    The status of the node from the TaskStatus enum.
    """

    start_time: Optional[datetime] = field(default_factory=datetime.now)
    """
    The starting time of the function, optional
    """

    end_time: Optional[datetime] = None
    """
    The ending time of the function, optional
    """

    reason: Optional[PlanFailure] = None
    """
    The reason of failure if the action failed.
    """

    plan: Plan = None
    """
    Reference to the plan to which this node belongs
    """

    result: Any = None
    """
    Result from the execution of this node
    """

    @property
    def parent(self) -> PlanNode:
        """
        The parent node of this node, None if this is the root node

        :return: The parent node
        """
        return list(self.plan.predecessors(self))[0] if list(self.plan.predecessors(self)) else None

    @property
    def children(self) -> List[PlanNode]:
        """
        All children nodes of this node

        :return:  A list of child nodes
        """
        return list(self.plan.successors(self))

    @property
    def recursive_children(self) -> List[PlanNode]:
        """
        Recursively lists all children and their children.

        :return: A list of all nodes below this node
        """
        return list(nx.descendants(self.plan, self))

    @property
    def subtree(self) -> Plan:
        """
        Creates a new plan with this node as the new root

        :return: A new plan
        """
        graph = nx.DiGraph()
        graph.add_nodes_from(self.plan.nodes)
        graph.add_edges_from(self.plan.edges)
        # The subgraph methods tries to create a new instance of the graph class it is give which in the case of Plan()
        # would fail because of the "root" param, that's the reason for this weird conversion.
        sub_grap = nx.subgraph(graph, [self] +  self.recursive_children)
        plan = Plan(self)
        plan.add_nodes_from(sub_grap.nodes)
        plan.add_edges_from(sub_grap.edges)
        return plan


    @property
    def all_parents(self) -> List[PlanNode]:
        """
        Returns all nodes above this node until the root node. The order is from this node to the root node.

        :return: A list of all nodes above this
        """
        return list(nx.ancestors(self.plan, self))

    @property
    def is_leaf(self) -> bool:
        """
        Returns True if this node is a leaf node

        :return: True if this node is a leaf node
        """
        return self.children == []

    def flattened_parameters(self):
        """
        The atomic types pf this node as dict

        :return: The flattened parameter
        """
        pass

    def __hash__(self):
        return id(self)

    def perform(self, *args, **kwargs):
        pass

    def interrupt(self):
        """
        Interrupts the execution of this node and all nodes below
        """
        self.status = TaskStatus.INTERRUPTED
        loginfo(f"Interrupted node: {str(self)}")
        if giskard.giskard_wrapper:
            giskard.giskard_wrapper.interrupt()

    def resume(self):
        """
        Resumes the execution of this node and all nodes below
        """
        self.status = TaskStatus.RUNNING

    def pause(self):
        """
        Suspends the execution of this node and all nodes below.
        """
        self.status = TaskStatus.SLEEPING


@dataclass
class DesignatorNode(PlanNode):
    designator_ref: Any = None
    """
    Reference to the Designator in this node
    """

    action: Optional[Any] = None
    """
    The action and that is performed or None if nothing was performed
    """

    kwargs: Dict[str, Any] = None
    """
    kwargs of the action in this node
    """

    def __hash__(self):
        return id(self)

    def __repr__(self, *args, **kwargs):
        return f"<{self.designator_ref.performable.__name__}>"

    def flattened_parameters(self) -> Dict[str, leaf_types]:
        """
        The atomic types of the parameters of this node as dict with paths as keys and the atomic type as value.
        This resolves the parameters to its type not the actual value.

        :return: The atomic types of this action
        """
        return self.designator_ref.performable.flattened_parameters()

    def flatten(self) -> Dict[str, leaf_types]:
        """
        Flattens the parameters of this node to a dict with the parameter as  key and the value as value.

        :return: A dict of the flattened parameters
        """
        return self.designator_ref.flatten()


@dataclass
class ActionNode(DesignatorNode):
    """
    A node in the plan representing an ActionDesignator description
    """
    action_iter: Iterator[ActionDescription] = None
    """
    Iterator over the current evaluation state of the ActionDesignator Description
    """

    action: ActionDescription = None

    def __hash__(self):
        return id(self)

    @managed_node
    def perform(self):
        """
        Performs this node by resolving the ActionDesignator description to the next resolution and then performing the
        result.

        :return: Return value of the resolved action node
        """
        if not self.action_iter:
            self.action_iter = iter(self.designator_ref)
        resolved_action = next(self.action_iter)
        kwargs = {key: resolved_action.__getattribute__(key) for key in self.designator_ref.kwargs.keys()}
        resolved_action_node = ResolvedActionNode(designator_ref=resolved_action, action=resolved_action.__class__,
                                                  kwargs=kwargs)
        self.plan.add_edge(self, resolved_action_node)

        return resolved_action_node.perform()

    def __repr__(self, *args, **kwargs):
        return f"<{self.designator_ref.performable.__name__}>"


@dataclass
class ResolvedActionNode(DesignatorNode):
    """
    A node representing a resolved ActionDesignator with fully specified parameters
    """
    designator_ref: ActionDescription = None

    action: ActionDescription = None

    def __hash__(self):
        return id(self)

    @managed_node
    def perform(self):
        """
        Performs this node by performing the resolved action designator in zit

        :return: The return value of the resolved ActionDesignator
        """
        return self.designator_ref.perform()

    def __repr__(self, *args, **kwargs):
        return f"<Resolved {self.designator_ref.__class__.__name__}>"



@dataclass
class MotionNode(DesignatorNode):
    """
    A node in the plan representing a MotionDesignator
    """
    designator_ref: BaseMotion = None
    """
    Reference to the MotionDesignator
    """

    def __hash__(self):
        return id(self)


    @managed_node
    def perform(self):
        """
        Performs this node by performing the respective MotionDesignator. Additionally, checks if one of the parents has
        the status INTERRUPTED and aborts the perform if that is the case.

        :return: The return value of the Motion Designator
        """
        return self.designator_ref.perform()

    def __repr__(self, *args, **kwargs):
        return f"<{self.designator_ref.__class__.__name__}>"

    def flatten(self):
        return {}

    def flattened_parameters(self):
        return {}


def with_plan(func: Callable) -> Callable:
    """
    Decorator which wrapps the decorated designator into a node, creates a new plan with the node as root and returns
    the plan.

    :param func: The decorator which should be inserted into a plan
    :return: A plan with the designator as root node
    """

    def wrapper(*args, **kwargs) -> Plan:
        designator = func(*args, **kwargs)
        if designator.__class__.__name__ == "PartialDesignator":
            node = ActionNode(designator_ref=designator, action=designator.performable, kwargs=designator.kwargs)
        else:
            kwargs = dict(inspect.signature(func).bind(*args, **kwargs).arguments)
            node = MotionNode(designator_ref=designator, action=designator.__class__, kwargs=kwargs)
        plan = Plan(root=node)
        if Plan.current_plan:
            Plan.current_plan.mount(plan, Plan.current_plan.current_node)
        return plan

    return wrapper



from owlready2 import Imp

from .dependencies import *
from .classes import *
from .individuals import *
from .object_properties import *

Floor.is_a = [PhysicalObject]

Milk.is_a = [Food]

Robot.is_a = [Agent]

Human.is_a = [Agent]

Cereal.is_a = [Food]

Kitchen.is_a = [Room]

Food.is_a = [PhysicalObject]

Apartment.is_a = [Room]

Container.is_a = [PhysicalObject]

Room.is_a = [Location, Container]

Cup.is_a = [Container]

Bowl.is_a = [Container]

DefaultPreferredGraspAlignment.is_a = [PreferredGraspAlignment, has_preferred_axis.some(NoAlignment),
                                       has_vertical_alignment.value(Falsy), has_rotated_gripper.value(Falsy),
                                       has_rim_grasp.value(Falsy)]

BowlPreferredGraspAlignment.is_a = [PreferredGraspAlignment, has_preferred_axis.value(NoAlignment),
                                    has_vertical_alignment.value(Truthy), has_rotated_gripper.value(Truthy),
                                    has_rim_grasp.value(Truthy)]

SpoonPreferredGraspAlignment.is_a = [PreferredGraspAlignment, has_preferred_axis.value(XAxis),
                                     has_vertical_alignment.value(Truthy), has_rotated_gripper.value(Falsy),
                                     has_rim_grasp.value(Falsy)]

CerealPreferredGraspAlignment.is_a = [PreferredGraspAlignment, has_preferred_axis.value(XAxis),
                                      has_vertical_alignment.value(Falsy), has_rotated_gripper.value(Falsy),
                                      has_rim_grasp.value(Falsy)]

Cabinet.is_a = [Container]

Drawer.is_a = [Container]

Spoon.is_a = [PhysicalObject]

ContinuousJoint.is_a = [HingeJoint]

HingeJoint.is_a = [MovableJoint]

FixedJoint.is_a = [Joint]

MovableJoint.is_a = [Joint]

PlanarJoint.is_a = [MovableJoint]

PrismaticJoint.is_a = [MovableJoint]

FloatingJoint.is_a = [MovableJoint]

RevoluteJoint.is_a = [HingeJoint]

Event.is_a = [Entity]

Entity.is_a = [Thing]

Action.is_a = [Event]

PhysicalTask.is_a = [Entity]

Supporter.is_a = [PhysicalObject]

SupportedObject.is_a = [PhysicalObject]

is_part_of.is_a = [ObjectProperty, TransitiveProperty, ReflexiveProperty]
is_part_of.domain = [PhysicalObject]
is_part_of.range = [PhysicalObject]

has_part.is_a = [ObjectProperty, TransitiveProperty, ReflexiveProperty]
has_part.domain = [PhysicalObject]
has_part.range = [PhysicalObject]

has_parent_link.is_a = [ObjectProperty]
has_parent_link.domain = [Joint]
has_parent_link.range = [PhysicalObject]

has_child_link.is_a = [ObjectProperty]
has_child_link.domain = [Joint]
has_child_link.range = [PhysicalObject]

contains.is_a = [ObjectProperty, TransitiveProperty]
contains.domain = [Base]
contains.range = [Base]
contains.inverse_property = is_contained_in

contains_object.is_a = [ObjectProperty, TransitiveProperty]
contains_object.domain = [Container]
contains_object.range = [PhysicalObject]
contains_object.inverse_property = is_physically_contained_in

is_contained_in.is_a = [ObjectProperty, TransitiveProperty]
is_contained_in.domain = [Base]
is_contained_in.range = [Base]

is_physically_contained_in.is_a = [ObjectProperty, TransitiveProperty]
is_physically_contained_in.domain = [PhysicalObject]
is_physically_contained_in.range = [Container]

supports.is_a = [ObjectProperty, TransitiveProperty]
supports.domain = [Supporter]
supports.range = [SupportedObject]
supports.inverse_property = is_supported_by

is_supported_by.is_a = [ObjectProperty, TransitiveProperty]
is_supported_by.domain = [SupportedObject]
is_supported_by.range = [Supporter]

for c in ontology.classes():
    c.has_preferred_alignment.append(DefaultPreferredGraspAlignment)

Bowl.has_preferred_alignment = [BowlPreferredGraspAlignment]
Spoon.has_preferred_alignment = [SpoonPreferredGraspAlignment]
Cereal.has_preferred_alignment = [CerealPreferredGraspAlignment]



import numpy as np

from pycrap.ontologies import PhysicalObject
from .object_descriptors.generic import ObjectDescription
from .tf_transformations import quaternion_from_euler
from typing_extensions import Tuple, List, Union, Dict, Iterable, Optional, Iterator

from .datastructures.enums import Arms
from .costmaps import Costmap
from .datastructures.pose import PoseStamped, TransformStamped, GraspDescription
from .datastructures.world import World
from .external_interfaces.ik import request_ik
from .failures import IKError, RobotInCollision
from .local_transformer import LocalTransformer
from .robot_description import RobotDescription
from .ros import logdebug
from .world_concepts.world_object import Object
from .world_reasoning import contact


class PoseGenerator(Iterable[PoseStamped]):
    """
    Creates pose candidates from a given costmap.
    The generator selects the highest values, amount is given by number_of_sample, and returns the corresponding
    positions.
    Orientations are calculated such that the Robot faces the center of the costmap.
    """

    current_orientation_generator = None
    """
    If no orientation generator is given, this generator is used to generate the orientation of the robot.
    """
    override_orientation_generator = None
    """
    Override the orientation generator with a custom generator, which will be used regardless of the current_orientation_generator.
    """

    def __init__(self, costmap: Costmap, number_of_samples=100, orientation_generator=None):
        """
        :param costmap: The costmap from which poses should be sampled.
        :param number_of_samples: The number of samples from the costmap that should be returned at max
        :param orientation_generator: function that generates an orientation given a position and the origin of the costmap
        """

        if not PoseGenerator.current_orientation_generator:
            PoseGenerator.current_orientation_generator = PoseGenerator.generate_orientation

        self.costmap = costmap
        self.number_of_samples = number_of_samples
        self.orientation_generator = orientation_generator if orientation_generator else PoseGenerator.current_orientation_generator
        if PoseGenerator.override_orientation_generator:
            self.orientation_generator = PoseGenerator.override_orientation_generator

    def __iter__(self) -> Iterator[PoseStamped]:
        """
        A generator that crates pose candidates from a given costmap. The generator
        selects the highest 100 values and returns the corresponding positions.
        Orientations are calculated such that the Robot faces the center of the costmap.

        :Yield: A tuple of position and orientation
        """

        # Determines how many positions should be sampled from the costmap
        if self.number_of_samples == -1 or self.number_of_samples > self.costmap.map.flatten().shape[0]:
            self.number_of_samples = self.costmap.map.flatten().shape[0]
        indices = np.argpartition(self.costmap.map.flatten(), -self.number_of_samples)[-self.number_of_samples:]
        indices = np.dstack(np.unravel_index(indices, self.costmap.map.shape)).reshape(self.number_of_samples, 2)

        height = self.costmap.map.shape[0]
        width = self.costmap.map.shape[1]
        center = np.array([height // 2, width // 2])
        for ind in indices:
            if self.costmap.map[ind[0]][ind[1]] == 0:
                continue
            # The position is calculated by creating a vector from the 2D position in the costmap (given by x and y)
            # and the center of the costmap (since this is the origin). This vector is then turned into a transformation
            # and muiltiplied with the transformation of the origin.
            vector_to_origin = (center - ind) * self.costmap.resolution
            point_to_origin = TransformStamped.from_list([*vector_to_origin, 0], frame="point", child_frame_id="origin")
            origin_to_map = ~self.costmap.origin.to_transform_stamped("origin")
            point_to_map = point_to_origin * origin_to_map
            map_to_point = ~point_to_map

            orientation = self.orientation_generator(map_to_point.translation.to_list(), self.costmap.origin)
            yield PoseStamped.from_list(map_to_point.translation.to_list(), orientation)

    @staticmethod
    def height_generator() -> float:
        pass

    @staticmethod
    def generate_orientation(position: List[float], origin: PoseStamped) -> List[float]:
        """
        This method generates the orientation for a given position in a costmap. The
        orientation is calculated such that the robot faces the origin of the costmap.
        This generation is done by simply calculating the arctan between the position,
        in the costmap, and the origin of the costmap.

        :param position: The position in the costmap. This position is already converted to the world coordinate frame.
        :param origin: The origin of the costmap. This is also the point which the robot should face.
        :return: A quaternion of the calculated orientation
        """
        angle = np.arctan2(position[1] - origin.position.y, position[0] - origin.position.x) + np.pi
        quaternion = list(quaternion_from_euler(0, 0, angle, axes="sxyz"))
        return quaternion


def visibility_validator(robot: Object,
                         object_or_pose: Union[Object, PoseStamped]) -> bool:
    """
    This method validates if the robot can see the target position from a given
    pose candidate. The target position can either be a position, in world coordinate
    system, or an object in the World. The validation is done by shooting a
    ray from the camera to the target position and checking that it does not collide
    with anything else.

    :param robot: The robot object for which this should be validated
    :param object_or_pose: The target position or object for which the pose candidate should be validated.
    :return: True if the target is visible for the robot, None in any other case.
    """
    world = robot.world
    robot_pose = robot.get_pose()

    if isinstance(object_or_pose, PoseStamped):
        gen_obj_desc = ObjectDescription("viz_object", [0, 0, 0], [0.02, 0.02, 0.02])
        obj = Object("viz_object", PhysicalObject, pose=object_or_pose, description=gen_obj_desc)
    else:
        obj = object_or_pose

    obj_id = obj.id

    camera_pose = robot.get_link_pose(RobotDescription.current_robot_description.get_camera_link())
    robot.set_pose(PoseStamped.from_list([100, 100, 0], [0, 0, 0, 1]))
    ray = world.ray_test(camera_pose.position.to_list(), obj.get_pose().position.to_list())
    robot.set_pose(robot_pose)
    if isinstance(object_or_pose, PoseStamped):
        world.remove_object(obj)

    return ray.obj_id == obj_id


def reachability_validator(robot: Object,
                           target_pose: PoseStamped,
                           arm: Arms,
                           allowed_collision: Dict[Object, List] = None
                           ) -> Optional[Dict[str, float]]:
    """
    This method validates if a target position is reachable for the robot.
    This is done by asking the ik solver if there is a valid solution if the
    robot stands at the position of the pose candidate. if there is a solution
    the validator returns True and False in any other case.

    :param robot: The robot object in the World for which the reachability should be validated.
    :param target_pose: The target pose for which the reachability should be validated.
    :param arm: The arm that should be checked for reachability.
    :param allowed_collision: dict of objects with which the robot is allowed to collide each object correlates
     to a list of links of which this object consists

    :return: The final joint states of the robot if the target pose is reachable without collision, None in any other case.
    """

    allowed_collision = allowed_collision or {}
    arm_chain = RobotDescription.current_robot_description.get_arm_chain(arm)
    allowed_collision[robot] = [link for link in arm_chain.end_effector.links]

    joint_state_before_ik = robot.get_positions_of_all_joints()
    try:
        pose, joint_states = request_ik(target_pose, robot, arm_chain.joints, arm_chain.end_effector.tool_frame)
        logdebug(f"Robot {arm.name} can reach target pose")
        robot.set_pose(pose)
        robot.set_multiple_joint_positions(joint_states)
        collision_check(robot, allowed_collision)
        logdebug(f"Robot is not in contact at target pose")

        robot.set_multiple_joint_positions(joint_state_before_ik)
        return joint_states

    except (IKError, RobotInCollision):
        logdebug(f"Robot {arm.name} cannot reach pose without collision")
        return None
    finally:
        robot.set_multiple_joint_positions(joint_state_before_ik)


def pose_sequence_reachability_validator(robot: Object,
                                         target_sequence: List[PoseStamped],
                                         arm: Arms,
                                         allowed_collision: Dict[Object, List] = None
                                         ) -> bool:
    """
    This method validates if a target sequence, if traversed in order, is reachable for the robot.
    This is done by asking the ik solver if there is a valid solution if the
    robot stands at the position of the pose candidate. If there is a solution
    the validator returns The arm that can reach the target position and None in any other case.

    :param robot: The robot object in the World for which the reachability should be validated.
    :param target_sequence: The target sequence of poses for which the reachability should be validated.
    :param arm: The arm that should be checked for reachability.
    :param allowed_collision: dict of objects with which the robot is allowed to collide each object correlates
     to a list of links of which this object consists

    :return: The arm that can reach the target position and None in any other case.
    """
    joint_state_before_ik = robot.get_positions_of_all_joints()
    for target_pose in target_sequence:
        joint_states = reachability_validator(robot, target_pose, arm, allowed_collision)
        if joint_states is None:
            robot.set_multiple_joint_positions(joint_state_before_ik)
            return False
        robot.set_multiple_joint_positions(joint_states)

    robot.set_multiple_joint_positions(joint_state_before_ik)
    return True


def collision_check(robot: Object, allowed_collision: Dict[Object, List]):
    """
    This method checks if a given robot collides with any object within the world
    which it is not allowed to collide with.
    This is done checking iterating over every object within the world and checking
    if the robot collides with it. Careful the floor will be ignored.
    If there is a collision with an object that was not within the allowed collision
    list the function will raise a RobotInCollision exception.

    :param robot: The robot object in the (Bullet)World where it should be checked if it collides with something
    :param allowed_collision: dict of objects with which the robot is allowed to collide each object correlates to a list of links of which this object consists

    :raises: RobotInCollision if the robot collides with an object it is not allowed to collide with.
    """
    allowed_robot_links = allowed_collision.get(robot, [])
    for obj in World.current_world.objects:
        if obj.name == "floor":
            continue
        in_contact, contact_links = contact(robot, obj, return_links=True)
        if not in_contact:
            continue

        allowed_links = allowed_collision.get(obj, [])
        if not all(link[0].name in allowed_robot_links or link[1].name in allowed_links for link in contact_links):
            raise RobotInCollision(f"Collision detected between {robot} and {obj}.")



import time

from pycram.testing import BulletWorldTestCase
from pycram.datastructures.enums import ObjectType
from pycram.datastructures.pose import PoseStamped
from pycram.datastructures.world import UseProspectionWorld
from pycram.world_concepts.world_object import Object
from pycrap.ontologies import Milk, Cereal


class TestAttachment(BulletWorldTestCase):

    def test_attach(self):
        self.milk.attach(self.robot)
        self.assertTrue(self.milk.attachments[self.robot])
        self.assertTrue(self.robot.attachments[self.milk])

    def test_detach(self):
        self.milk.attach(self.robot)
        self.milk.detach(self.robot)
        self.assertTrue(self.robot not in self.milk.attachments)
        self.assertTrue(self.milk not in self.robot.attachments)

    def test_detach_sync_in_prospection_world(self):
        self.milk.attach(self.robot)
        with UseProspectionWorld():
            pass
        self.milk.detach(self.robot)
        with UseProspectionWorld():
            pass
        self.assertTrue(self.milk not in self.robot.attachments)
        self.assertTrue(self.robot not in self.milk.attachments)
        prospection_milk = self.world.get_prospection_object_for_object(self.milk)
        prospection_robot = self.world.get_prospection_object_for_object(self.robot)
        self.assertTrue(prospection_milk not in prospection_robot.attachments)
        self.assertTrue(prospection_robot not in prospection_milk.attachments)

    def test_attachment_behavior(self):
        self.robot.attach(self.milk)

        milk_pos = self.milk.get_position()
        rob_pos = self.robot.get_position()

        rob_pos.x += 1
        self.robot.set_position(rob_pos)

        new_milk_pos = self.milk.get_position()
        self.assertEqual(new_milk_pos.x, milk_pos.x + 1)

    def test_detachment_behavior(self):
        self.robot.attach(self.milk)

        milk_pos = self.milk.get_position()
        rob_pos = self.robot.get_position()

        self.robot.detach(self.milk)
        rob_pos.x += 1
        self.robot.set_position(rob_pos)

        new_milk_pos = self.milk.get_position()
        self.assertEqual(new_milk_pos.x, milk_pos.x)

    def test_prospection_object_attachments_not_changed_with_real_object(self):
        milk_2 = Object("milk_2", Milk, "milk.stl", pose=PoseStamped.from_list([1.3, 1, 0.9]))
        cereal_2 = Object("cereal_2", Cereal, "breakfast_cereal.stl",
                          pose=PoseStamped.from_list([1.3, 0.7, 0.95]))
        time.sleep(0.05)
        milk_2.attach(cereal_2)
        time.sleep(0.05)
        with UseProspectionWorld():
            prospection_milk = self.world.get_prospection_object_for_object(milk_2)
            # self.assertTrue(cereal_2 not in prospection_milk.attachments)
            prospection_cereal = self.world.get_prospection_object_for_object(cereal_2)
            # self.assertTrue(prospection_cereal in prospection_milk.attachments)
            self.assertTrue(prospection_cereal in prospection_milk.attachments.keys())

            # Assert that when prospection object is moved, the real object is not moved
            prospection_milk_pos = prospection_milk.get_position()
            cereal_pos = cereal_2.get_position()
            estimated_prospection_cereal_pos = prospection_cereal.get_position()
            estimated_prospection_cereal_pos.x += 1

            # Move prospection milk object
            prospection_milk_pos.x += 1
            prospection_milk.set_position(prospection_milk_pos)

            # Prospection cereal should move since it is attached to prospection milk
            new_prospection_cereal_pose = prospection_cereal.get_position()
            self.assertAlmostEqual(new_prospection_cereal_pose.x, estimated_prospection_cereal_pos.x, delta=0.01)

            # Also Real cereal object should not move since it is not affected by prospection milk
            new_cereal_pos = cereal_2.get_position()
            assumed_cereal_pos = cereal_pos
            self.assertTrue(new_cereal_pos == assumed_cereal_pos)

        self.world.remove_object(milk_2)
        self.world.remove_object(cereal_2)

    def test_attaching_to_robot_and_moving(self):
        self.robot.attach(self.milk)
        milk_pos = self.milk.get_position()
        rob_pos = self.robot.get_position()

        rob_pos.x += 1
        self.robot.set_position(rob_pos)

        new_milk_pos = self.milk.get_position()
        self.assertEqual(new_milk_pos.x, milk_pos.x + 1)



def create_publisher(topic, msg_type, queue_size=10, latch: bool = False) -> None:
    """
    Create a ROS publisher.

    :param topic: The name of the topic to publish to.
    :param msg_type: The type of message to publish.
    :param queue_size: The size of the message queue.
    :param latch: Whether to latch the last message sent.
    :return: A ROS publisher object.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to create the publisher.
    raise NotImplementedError("ROS is not available in this environment. ")


from sqlalchemy import  types


class StringType(types.TypeDecorator):
    """
    This type represents a physical object type.
    The database representation of this is a string while the in memory type is the instance of PhysicalObject.
    """
    cache_ok = True
    impl = types.String

    def process_bind_param(self, value, dialect):
        return value.__class__.__name__

    def copy(self, **kw):
        return self.__class__(**kw)


from __future__ import annotations

import math
import numpy as np
from abc import abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from time import sleep
from typing import Optional, Union, Iterable, Tuple

from typing_extensions import Any

from .. import utils
from .motion_designator import MoveTCPMotion
from ..datastructures.enums import Arms, Grasp, AxisIdentifier
from ..datastructures.partial_designator import PartialDesignator
from ..datastructures.pose import PoseStamped
from ..datastructures.world import World
from ..designator import ActionDescription
from ..has_parameters import has_parameters
from ..local_transformer import LocalTransformer
from ..plan import with_plan
from ..robot_description import RobotDescription
from ..world_concepts.world_object import Object


@dataclass
class GAP(ActionDescription):
    object_: Object
    tool: Object
    arm: Arms
    technique: Optional[str] = None

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        pass

    @abstractmethod
    def plan(self) -> None:
        ...


@has_parameters
@dataclass
class MixingAction(GAP):
    def plan(self) -> None:
        lt = LocalTransformer()
        obj = self.object_
        pose = lt.transform_to_object_frame(obj.pose, obj)
        height_offset = obj.size[2] + 0.05

        for t in range(20):  # 2 * steps, with steps=10
            p = pose.copy()
            r, a, h = 0.0035 * t, math.radians(30) * t, 0.001 * t
            p.pose.position.x += r * math.cos(a)
            p.pose.position.y += r * math.sin(a)
            p.pose.position.z += h

            spiral = lt.transform_pose(p, "map")
            spiral.pose.position.z += height_offset
            World.current_world.add_vis_axis(spiral)
            MoveTCPMotion(spiral, self.arm).perform()

        World.current_world.remove_vis_axis()

    @classmethod
    @with_plan
    def description(cls, object_: Union[Iterable[Object], Object], tool: Union[Iterable[Object], Object],
                    arm: Optional[Union[Iterable[Arms], Arms]] = None,
                    technique: Optional[Union[Iterable[str], str]] = None):
        return PartialDesignator(cls, object_=object_, tool=tool, arm=arm, technique=technique)


@dataclass
class CuttingAction(GAP):
    slice_thickness: Optional[float] = 0.03

    def plan(self) -> None:
        if self.technique is None:
            self.technique = "Slicing"
        lt = LocalTransformer()
        obj = self.object_
        tool = self.tool
        pose = lt.transform_to_object_frame(obj.pose, obj)

        height = obj.size[2]
        length = max(obj.size[0], obj.size[1])
        length_tool = max(tool.size[0], tool.size[1])

        num_slices, start_offset = self.calculate_slices(length)
        slice_coordinates = [start_offset + i * self.slice_thickness for i in range(num_slices)]

        slice_poses = []
        for x in slice_coordinates:
            tmp_pose = pose.copy()
            tmp_pose.pose.position.x = x
            slice_poses.append(tmp_pose)

        for slice_pose in slice_poses:
            pose_a = obj.pose
            pose_b = World.robot.pose
            angle, angle_y = self.get_rotation_offset_from_axis_preference(pose_a, pose_b)
            direction = 1 if angle_y >= 0 else -1
            slice_pose.pose.position.y += direction * (length_tool / 2)

            new_pose = self.perpendicular_pose(slice_pose=slice_pose, angle=angle)
            final_pose = lt.transform_pose(new_pose, "map")

            print(final_pose)
            World.current_world.add_vis_axis(final_pose)

            lift_pose = new_pose.copy()
            lift_pose.pose.position.z += height

    @classmethod
    @with_plan
    def description(cls, object_: Union[Iterable[Object], Object], tool: Union[Iterable[Object], Object],
                    arm: Optional[Union[Iterable[Arms], Arms]] = None,
                    technique: Optional[Union[Iterable[str], str]] = None, slice_thickness: Optional[float] = 0.03):
        return PartialDesignator(cls, object_=object_, tool=tool, arm=arm, technique=technique,
                                 slice_thickness=slice_thickness)

    def calculate_slices(self, obj_length):
        if self.technique == 'Halving':
            return 1, 0
        if self.technique in ['Cutting Action', 'Sawing', 'Paring', 'Cutting', 'Carving', 'Slicing']:
            num_slices = int(obj_length // self.slice_thickness)
            start_offset = (-obj_length / 2) + (self.slice_thickness / 2)
            return num_slices, start_offset
        return 0, 0

    @staticmethod
    def perpendicular_pose(slice_pose, angle) -> PoseStamped:
        pose_rotated = slice_pose
        rotation_quaternion = utils.axis_angle_to_quaternion([0, 0, 1], angle)
        pose_rotated.rotate_by_quaternion(rotation_quaternion)
        return pose_rotated

    @staticmethod
    def get_rotation_offset_from_axis_preference(pose_a, pose_b: PoseStamped) -> Tuple[int, float]:
        """
        Compute a discrete rotation offset (-90 or 90 degrees) to align this pose's local axes with the direction
        toward a target pose, based on which axis (X or Y) is more aligned.

        :param pose_a: The source pose.
        :param pose_b: The target pose to align with.
        :return: Tuple of (rotation offset in degrees, signed angle difference in radians for Y axis).
        """
        fx, ax = pose_a.is_facing_2d_axis(pose_b, axis=AxisIdentifier.X)
        fy, ay = pose_a.is_facing_2d_axis(pose_b, axis=AxisIdentifier.Y)

        return (-90 if abs(ax) > abs(ay) else 90), ay




@dataclass
class PouringAction(GAP):
    angle: Optional[float] = 90

    def plan(self) -> None:
        lt = LocalTransformer()
        gripper_frame = World.robot.get_link_tf_frame("base_link")
        grasp_rot = RobotDescription.current_robot_description.get_arm_chain(self.arm).end_effector.get_grasp(
            Grasp.FRONT, None, False)

        pose = lt.transform_pose(self.object_.pose, gripper_frame)
        pose.pose.position.x += 0.009
        pose.pose.position.y -= 0.125
        pose.pose.position.z += 0.17

        pose = lt.transform_pose(lt.transform_pose(pose, "map"), gripper_frame)
        pose.orientation = grasp_rot
        pose = lt.transform_pose(pose, "map")

        World.current_world.add_vis_axis(pose)
        # MoveTCPMotion(pose, self.arm, allow_gripper_collision=False, movement_type=MovementType.CARTESIAN).perform()

        pour_pose = pose.copy()
        pour_pose.rotate_by_quaternion(utils.axis_angle_to_quaternion([1, 0, 0], -self.angle))
        World.current_world.add_vis_axis(pour_pose)

        # MoveTCPMotion(pour_pose, self.arm, allow_gripper_collision=False,movement_type=MovementType.CARTESIAN).perform()
        sleep(3)
        # MoveTCPMotion(pose, self.arm, allow_gripper_collision=False, movement_type=MovementType.CARTESIAN).perform()

        World.current_world.remove_vis_axis()

    @classmethod
    @with_plan
    def description(cls, object_: Union[Iterable[Object], Object], tool: Union[Iterable[Object], Object],
                    arm: Optional[Union[Iterable[Arms], Arms]] = None,
                    technique: Optional[Union[Iterable[str], str]] = None,
                    angle: Optional[Union[Iterable[float], float]] = 90):
        return PartialDesignator(cls, object_=object_, tool=tool, arm=arm, technique=technique, angle=angle)


PouringActionDescription = PouringAction.description
MixingActionDescription = MixingAction.description
CuttingActionDescription = CuttingAction.description



import json
import threading

import sys

from ..ros import Time
from ..ros import logwarn, loginfo_once, loginfo
from ..ros import get_node_names

from ..datastructures.enums import JointType, ObjectType, Arms
from ..datastructures.pose import PoseStamped
from ..datastructures.world import World
from ..datastructures.dataclasses import MeshVisualShape
from ..ros import get_service_proxy
from ..world_concepts.world_object import Object
from ..robot_description import RobotDescription

from typing_extensions import List, Dict, Callable, Optional
from threading import Lock, RLock
from pycram.ros import logging as log



giskard_wrapper = None
giskard_update_service = None
is_init = False

number_of_par_goals = 0
giskard_lock = Lock()
giskard_rlock = RLock()
with giskard_rlock:
    par_threads = {}
    par_motion_goal = {}


def thread_safe(func: Callable) -> Callable:
    """
    Adds thread safety to a function via a decorator. This uses the giskard_lock

    :param func: Function that should be thread safe
    :return: A function with thread safety
    """

    def wrapper(*args, **kwargs):
        with giskard_rlock:
            return func(*args, **kwargs)

    return wrapper


def init_giskard_interface(func: Callable) -> Callable:
    """
    Checks if the ROS messages are available and if giskard is running, if that is the case the interface will be
    initialized.

    :param func: Function this decorator should be wrapping
    :return: A callable function which initializes the interface and then calls the wrapped function
    """

    def wrapper(*args, **kwargs):

        from giskardpy.python_interface.old_python_interface import OldGiskardWrapper as GiskardWrapper
        from giskard_msgs.msg import WorldBody, MoveResult, CollisionEntry
        from geometry_msgs.msg import PoseStamped as ROSPoseStamped, PointStamped, QuaternionStamped, Vector3Stamped

        global giskard_wrapper
        global giskard_update_service
        global is_init
        if is_init and "/giskard" in get_node_names():
            return func(*args, **kwargs)
        elif is_init and "/giskard" not in get_node_names():
            logwarn("Giskard node is not available anymore, could not initialize giskard interface")
            is_init = False
            giskard_wrapper = None
            return

        if "giskard_msgs" not in sys.modules:
            logwarn("Could not initialize the Giskard interface since the giskard_msgs are not imported")
            return

        if "/giskard" in get_node_names():
            giskard_wrapper = GiskardWrapper()
            loginfo_once("Successfully initialized Giskard interface")
            is_init = True
        else:
            logwarn("Giskard is not running, could not initialize Giskard interface")
            return
        return func(*args, **kwargs)
    return wrapper


# Believe state management between pycram and giskard


@init_giskard_interface
def initial_adding_objects() -> None:
    """
    Adds object that are loaded in the World to the Giskard belief state, if they are not present at the moment.
    """
    groups = giskard_wrapper.get_group_names()
    for obj in World.current_world.objects:
        if obj is World.robot or obj is World.current_world.get_prospection_object_for_object(World.robot):
            continue
        name = obj.name
        if name not in groups:
            spawn_object(obj)


@init_giskard_interface
def removing_of_objects() -> None:
    """
    Removes objects that are present in the Giskard belief state but not in the World from the Giskard belief state.
    """
    groups = giskard_wrapper.get_group_names()
    object_names = list(
        map(lambda obj: object_names.name, World.current_world.objects))
    diff = list(set(groups) - set(object_names))
    for grp in diff:
        giskard_wrapper.remove_group(grp)


@init_giskard_interface
def sync_worlds(projection: bool = False) -> None:
    """
    Synchronizes the World and the Giskard belief state, this includes adding and removing objects to the Giskard
    belief state such that it matches the objects present in the World and moving the robot to the position it is
    currently at in the World.

    :param projection: Whether the sync in projection mode or reality.
    """
    add_gripper_groups()
    world_object_names = set()
    for obj in World.current_world.objects:
        if obj.name != RobotDescription.current_robot_description.name and obj.obj_type != ObjectType.ROBOT and len(obj.link_name_to_id) != 1:
            world_object_names.add(obj.name)
        if obj.name == RobotDescription.current_robot_description.name or obj.obj_type == ObjectType.ROBOT:
            joint_config = obj.get_positions_of_all_joints()
            non_fixed_joints = list(filter(lambda joint: joint.type != JointType.FIXED and not joint.is_virtual,
                                           obj.joints.values()))
            joint_config_filtered = {joint.name: joint_config[joint.name] for joint in non_fixed_joints}

            if projection:
                giskard_wrapper.monitors.add_set_seed_configuration(joint_config_filtered,
                                                                    RobotDescription.current_robot_description.name)
                giskard_wrapper.monitors.add_set_seed_odometry(obj.get_pose().ros_message(),
                                                               RobotDescription.current_robot_description.name)
    giskard_object_names = set(giskard_wrapper.get_group_names())
    robot_name = {RobotDescription.current_robot_description.name}
    if not world_object_names.union(robot_name).issubset(giskard_object_names):
        giskard_wrapper.clear_world()
    initial_adding_objects()


@init_giskard_interface
def update_pose(object: Object) -> 'UpdateWorldResponse':
    """
    Sends an update message to giskard to update the object position. Might not work when working on the real robot just
    in standalone mode.

    :param object: Object that should be updated
    :return: An UpdateWorldResponse
    """
    return giskard_wrapper.update_group_pose(object.name)


@init_giskard_interface
def spawn_object(object: Object) -> None:
    """
    Spawns a World Object in the giskard belief state.

    :param object: World object that should be spawned
    """
    if len(object.link_name_to_id) == 1:
        geometry = object.get_link_geometry(object.root_link.name)
        if isinstance(geometry, MeshVisualShape):
            filename = geometry.file_name
            spawn_mesh(object.name, filename, object.get_pose())
    else:
        ww = spawn_urdf(object.name, object.path, object.get_pose())
        log.loginfo("GiskardSpawnURDF Return value: {} ObjectName:{}".format(ww,object.name))


@init_giskard_interface
def remove_object(object: Object) -> 'UpdateWorldResponse':
    """
    Removes an object from the giskard belief state.

    :param object: The World Object that should be removed
    """
    return giskard_wrapper.remove_group(object.name)


@init_giskard_interface
def spawn_urdf(name: str, urdf_path: str, pose: PoseStamped) -> 'UpdateWorldResponse':
    """
    Spawns an URDF in giskard's belief state.

    :param name: Name of the URDF
    :param urdf_path: Path to the URDF file
    :param pose: Pose in which the URDF should be spawned
    :return: An UpdateWorldResponse message
    """
    urdf_string = ""
    with open(urdf_path) as f:
        urdf_string = f.read()

    return giskard_wrapper.add_urdf(name, urdf_string, pose)


@init_giskard_interface
def spawn_mesh(name: str, path: str, pose: 'PoseStamped') -> 'UpdateWorldResponse':
    """
    Spawns a mesh into giskard's belief state

    :param name: Name of the mesh
    :param path: Path to the mesh file
    :param pose: Pose in which the mesh should be spawned
    :return: An UpdateWorldResponse message
    """
    return giskard_wrapper.add_mesh(name, path, pose)


# Sending Goals to Giskard

@thread_safe
def _manage_par_motion_goals(goal_func, *args) -> Optional['MoveResult']:
    """
    Manages multiple goals that should be executed in parallel. The current sequence of motion goals is saved and the
    parallel motion goal is loaded if there is one, then the new motion goal given by ``goal_func`` is added to the
    parallel motion goal. If this was the last motion goal for the parallel motion goal it is then executed.

    :param goal_func: Function which adds a new motion goal to the giskard_wrapper
    :param args: Arguments for the ``goal_func`` function
    :return: MoveResult of the execution if there was an execution, True if a new motion goal was added to the giskard_wrapper and None in any other case
    """
    # key is the instance of the parallel language element, value is a list of threads that should be executed in
    # parallel
    for key, value in par_threads.items():
        # if the current thread is in the list of threads that should be executed in parallel backup the current list
        # of motion goals and monitors
        if threading.get_ident() in value:
            tmp_goals = giskard_wrapper.motion_goals.get_goals()
            tmp_monitors = giskard_wrapper.monitors.get_monitors()

            if key in par_motion_goal.keys():
                # giskard_wrapper.cmd_seq = par_motion_goal[key]
                giskard_wrapper.motion_goals._goals = par_motion_goal[key][0]
                giskard_wrapper.monitors._monitors = par_motion_goal[key][1]
            else:
                giskard_wrapper.clear_motion_goals_and_monitors()

            goal_func(*args)

            # Check if there are multiple constraints that use the same joint, if this is the case the
            used_joints = set()
            for cmd in giskard_wrapper.motion_goals.get_goals():
                par_value_pair = json.loads(cmd.kwargs)
                if "tip_link" in par_value_pair.keys() and "root_link" in par_value_pair.keys():
                    if par_value_pair["tip_link"] == RobotDescription.current_robot_description.base_link:
                        continue
                    chain = World.robot.description.get_chain(par_value_pair["root_link"],
                                                              par_value_pair["tip_link"])
                    if set(chain).intersection(used_joints) != set():
                        giskard_wrapper.motion_goals._goals = tmp_goals
                        giskard_wrapper.monitors._monitors = tmp_monitors
                        raise AttributeError(
                            f"The joint(s) {set(chain).intersection(used_joints)} is used by multiple Designators")
                    else:
                        [used_joints.add(joint) for joint in chain]

                elif "goal_state" in par_value_pair.keys():
                    if set(par_value_pair["goal_state"].keys()).intersection(used_joints) != set():
                        giskard_wrapper.motion_goals._goals = tmp_goals
                        giskard_wrapper.monitors._monitors = tmp_monitors
                        raise AttributeError(
                            f"The joint(s) {set(par_value_pair['goal_state'].keys()).intersection(used_joints)} is used by multiple Designators")
                    else:
                        [used_joints.add(joint) for joint in par_value_pair["goal_state"].keys()]

            par_threads[key].remove(threading.get_ident())
            # If this is the last thread that should be executed in parallel, execute the complete sequence of motion
            # goals
            if len(par_threads[key]) == 0:
                if key in par_motion_goal.keys():
                    del par_motion_goal[key]
                del par_threads[key]
                # giskard_wrapper.add_default_end_motion_conditions()
                res = giskard_wrapper.execute()
                giskard_wrapper.motion_goals._goals = tmp_goals
                giskard_wrapper.monitors._monitors = tmp_monitors
                return res
            # If there are still threads that should be executed in parallel, save the current state of motion goals and
            # monitors.
            else:
                par_motion_goal[key] = [giskard_wrapper.motion_goals.get_goals(),
                                        giskard_wrapper.monitors.get_monitors()]
                giskard_wrapper.motion_goals._goals = tmp_goals
                giskard_wrapper.monitors._monitors = tmp_monitors
                return True


@init_giskard_interface
@thread_safe
def achieve_joint_goal(goal_poses: Dict[str, float]) -> 'MoveResult':
    """
    Takes a dictionary of joint position that should be achieved, the keys in the dictionary are the joint names and
    values are the goal joint positions.

    :param goal_poses: Dictionary with joint names and position goals
    :return: MoveResult message for this goal
    """
    set_joint_goal(goal_poses)
    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def set_joint_goal(goal_poses: Dict[str, float]) -> None:
    """
    Takes a dictionary of joint position that should be achieved, the keys in the dictionary are the joint names and
    values are the goal joint positions.

    :param goal_poses: Dictionary with joint names and position goals
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_joint_goal, goal_poses)
    if par_return:
        return par_return

    giskard_wrapper.set_joint_goal(goal_poses)


@init_giskard_interface
@thread_safe
def achieve_cartesian_goal(goal_pose: 'PoseStamped', tip_link: str, root_link: str,
                           position_threshold: float = 0.02,
                           orientation_threshold: float = 0.02,
                           use_monitor: bool = True,
                           grippers_that_can_collide: Optional[Arms] = None) -> 'MoveResult':
    """
    Takes a cartesian position and tries to move the tip_link to this position using the chain defined by
    tip_link and root_link.

    :param goal_pose: The position which should be achieved with tip_link
    :param tip_link: The end link of the chain as well as the link which should achieve the goal_pose
    :param root_link: The starting link of the chain which should be used to achieve this goal
    :param position_threshold: Position distance at which the goal is successfully reached
    :param orientation_threshold: Orientation distance at which the goal is successfully reached
    :param use_monitor: Whether to use a monitor for this goal or not.
    :param grippers_that_can_collide: The gripper(s) that should be allowed to collide.
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_cart_goal, goal_pose.ros_message(),
                                          tip_link, root_link)
    if par_return:
        return par_return

    cart_monitor1 = None
    if use_monitor:
        cart_monitor1 = giskard_wrapper.monitors.add_cartesian_pose(root_link=root_link, tip_link=tip_link,
                                                                    goal_pose=goal_pose.ros_message(),
                                                                    position_threshold=position_threshold, orientation_threshold=orientation_threshold,
                                                                    name='cart goal 1')
        end_monitor = giskard_wrapper.monitors.add_local_minimum_reached(start_condition=cart_monitor1)

    giskard_wrapper.motion_goals.add_cartesian_pose(name='g1', root_link=root_link, tip_link=tip_link,
                                                    goal_pose=goal_pose.ros_message(),
                                                    end_condition=cart_monitor1)

    if use_monitor:
        giskard_wrapper.monitors.add_end_motion(start_condition=end_monitor)

    giskard_wrapper.motion_goals.avoid_all_collisions()
    if grippers_that_can_collide is not None:
        allow_gripper_collision(grippers_that_can_collide)

    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def achieve_straight_cartesian_goal(goal_pose: 'PoseStamped', tip_link: str,
                                    root_link: str,
                                    grippers_that_can_collide: Optional[Arms] = None) -> 'MoveResult':
    """
    Takes a cartesian position and tries to move the tip_link to this position in a straight line, using the chain
    defined by tip_link and root_link.

    :param goal_pose: The position which should be achieved with tip_link
    :param tip_link: The end link of the chain as well as the link which should achieve the goal_pose
    :param root_link: The starting link of the chain which should be used to achieve this goal
    :param grippers_that_can_collide: The gripper(s) that should be allowed to collide.
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_straight_cart_goal, goal_pose.ros_message(),
                                          tip_link, root_link)
    if par_return:
        return par_return

    giskard_wrapper.avoid_all_collisions()
    if grippers_that_can_collide is not None:
        allow_gripper_collision(grippers_that_can_collide)
    giskard_wrapper.set_straight_cart_goal(goal_pose.ros_message(), tip_link, root_link)
    # giskard_wrapper.add_default_end_motion_conditions()
    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def achieve_translation_goal(goal_point: List[float], tip_link: str, root_link: str) -> 'MoveResult':
    """
    Tries to move the tip_link to the position defined by goal_point using the chain defined by root_link and
    tip_link. Since goal_point only defines the position but no rotation, rotation is not taken into account.

    :param goal_point: The goal position of the tip_link
    :param tip_link: The link which should be moved to goal_point as well as the end of the used chain
    :param root_link: The start link of the chain
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_translation_goal, make_point_stamped(goal_point),
                                          tip_link, root_link)
    if par_return:
        return par_return

    giskard_wrapper.set_translation_goal(make_point_stamped(goal_point), tip_link, root_link)
    # giskard_wrapper.add_default_end_motion_conditions()
    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def achieve_straight_translation_goal(goal_point: List[float], tip_link: str, root_link: str) -> 'MoveResult':
    """
    Tries to move the tip_link to the position defined by goal_point in a straight line, using the chain defined by
    root_link and tip_link. Since goal_point only defines the position but no rotation, rotation is not taken into account.

    :param goal_point: The goal position of the tip_link
    :param tip_link: The link which should be moved to goal_point as well as the end of the used chain
    :param root_link: The start link of the chain
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_straight_translation_goal,
                                          make_point_stamped(goal_point),
                                          tip_link, root_link)
    if par_return:
        return par_return

    giskard_wrapper.set_straight_translation_goal(make_point_stamped(goal_point), tip_link, root_link)
    # giskard_wrapper.add_default_end_motion_conditions()
    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def achieve_rotation_goal(quat: List[float], tip_link: str, root_link: str) -> 'MoveResult':
    """
    Tries to bring the tip link into the rotation defined by quat using the chain defined by root_link and
    tip_link.

    :param quat: The rotation that should be achieved, given as a quaternion
    :param tip_link: The link that should be in the rotation defined by quat
    :param root_link: The start link of the chain
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_rotation_goal, make_quaternion_stamped(quat),
                                          tip_link, root_link)
    if par_threads:
        return par_return

    giskard_wrapper.set_rotation_goal(make_quaternion_stamped(quat), tip_link, root_link)
    # giskard_wrapper.add_default_end_motion_conditions()
    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def achieve_align_planes_goal(goal_normal: List[float], tip_link: str, tip_normal: List[float],
                              root_link: str) -> 'MoveResult':
    """
    Tries to align the plane defined by tip normal with goal_normal using the chain between root_link and
    tip_link.

    :param goal_normal: The goal plane, given as a list of XYZ
    :param tip_link: The end link of the chain that should be used.
    :param tip_normal: The plane that should be aligned with goal_normal, given as a list of XYZ
    :param root_link: The starting link of the chain that should be used.
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_align_planes_goal, make_vector_stamped(goal_normal),
                                          tip_link, make_vector_stamped(tip_normal), root_link)
    if par_return:
        return par_return

    giskard_wrapper.set_align_planes_goal(make_vector_stamped(goal_normal), tip_link,
                                          make_vector_stamped(tip_normal),
                                          root_link)
    # giskard_wrapper.add_default_end_motion_conditions()
    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def achieve_open_container_goal(tip_link: str, environment_link: str) -> 'MoveResult':
    """
    Tries to open a container in an environment, this only works if the container was added as a URDF. This goal assumes
    that the handle was already grasped. Can only handle container with 1 DOF

    :param tip_link: The End effector that should open the container
    :param environment_link: The name of the handle for this container.
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_open_container_goal, tip_link, environment_link)
    if par_return:
        return par_return
    giskard_wrapper.set_open_container_goal(tip_link, environment_link)
    # giskard_wrapper.add_default_end_motion_conditions()
    return giskard_wrapper.execute()


@init_giskard_interface
@thread_safe
def achieve_close_container_goal(tip_link: str, environment_link: str) -> 'MoveResult':
    """
    Tries to close a container, this only works if the container was added as a URDF. Assumes that the handle of the
    container was already grasped. Can only handle container with 1 DOF.

    :param tip_link: Link name that should be used to close the container.
    :param environment_link: Name of the handle
    :return: MoveResult message for this goal
    """
    sync_worlds()
    par_return = _manage_par_motion_goals(giskard_wrapper.set_close_container_goal, tip_link, environment_link)
    if par_return:
        return par_return

    giskard_wrapper.set_close_container_goal(tip_link, environment_link)
    # giskard_wrapper.add_default_end_motion_conditions()
    return giskard_wrapper.execute()


@init_giskard_interface
def achieve_cartesian_waypoints_goal(waypoints: List['PoseStamped'], tip_link: str,
                                     root_link: str, enforce_final_orientation: bool = True) -> 'MoveResult':
    """
        Tries to achieve each waypoint in the given sequence of waypoints.
        If :param enforce_final_orientation is False, each waypoint needs a corresponding orientation. If it is True only
        the last waypoint needs to have an orientation.

        :param waypoints: The sequence of waypoints as poses to achieve.
        :param tip_link: The endeffector link of the chain that should be used.
        :param root_link: The root link of the chain that should be used.
        :param enforce_final_orientation: If true, only achieve the orientation of the last waypoint. If false, achieve the orientation of each waypoint.
        :return: MoveResult message for this goal
        """
    old_position_monitor = None
    old_orientation_monitor = None

    for i, waypoint in enumerate(waypoints):
        point = make_point_stamped(waypoint.position_as_list())
        orientation = make_quaternion_stamped(waypoint.orientation_as_list())
        start_condition = '' if not old_position_monitor else old_position_monitor

        # -------- Monitor Logic ------------
        if not enforce_final_orientation or (enforce_final_orientation and i == len(waypoints) - 1):
            if not enforce_final_orientation:
                start_condition = '' if not old_orientation_monitor else f'{old_orientation_monitor} and {old_position_monitor}'
            orientation_monitor = giskard_wrapper.monitors.add_cartesian_orientation(goal_orientation=orientation,
                                                                     tip_link=tip_link, root_link=root_link,
                                                                     start_condition=start_condition,
                                                                     name=str(id(waypoint)) + 'orientation')
            old_orientation_monitor = orientation_monitor

        # in all cases a position monitor is needed for each waypoint
        position_monitor = giskard_wrapper.monitors.add_cartesian_position(goal_point=point, tip_link=tip_link,
                                                            root_link=root_link,
                                                            start_condition=start_condition, name=str(id(waypoint)),
                                                            threshold=0.01 + (0.01 * (len(waypoints) - 1 - i)))
        # -------- Task Logic ---------------
        task_end_condition = position_monitor
        if not enforce_final_orientation or (enforce_final_orientation and i == len(waypoints) - 1):
            task_end_condition = f'{orientation_monitor} and {position_monitor}'
            giskard_wrapper.motion_goals.add_cartesian_orientation(goal_orientation=orientation,
                                                                       tip_link=tip_link, root_link=root_link,
                                                                       end_condition=task_end_condition,
                                                                       start_condition=start_condition,
                                                                       name=str(id(waypoint)) + 'orientation')

        # in all cases a position goal is needed for each waypoint
        giskard_wrapper.motion_goals.add_cartesian_position(goal_point=point, tip_link=tip_link,
                                                            root_link=root_link,
                                                            end_condition=task_end_condition, start_condition=start_condition,
                                                            name=str(id(waypoint)))


        old_position_monitor = position_monitor


    giskard_wrapper.monitors.add_end_motion(start_condition=f'{old_position_monitor} and {old_orientation_monitor}')
    giskard_wrapper.monitors.add_max_trajectory_length(30)
    giskard_wrapper.execute(add_default=False)


# Projection Goals


@init_giskard_interface
def projection_cartesian_goal(goal_pose: 'PoseStamped', tip_link: str, root_link: str) -> 'MoveResult':
    """
    Tries to move the tip_link to the position defined by goal_pose using the chain defined by tip_link and root_link.
    The goal_pose is projected to the closest point on the robot's workspace.

    :param goal_pose: The position which should be achieved with tip_link
    :param tip_link: The end link of the chain as well as the link which should achieve the goal_pose
    :param root_link: The starting link of the chain which should be used to achieve this goal
    :return: MoveResult message for this goal
    """
    sync_worlds(projection=True)
    giskard_wrapper.set_cart_goal(goal_pose.ros_message(), tip_link, root_link)
    return giskard_wrapper.projection()


@init_giskard_interface
def projection_cartesian_goal_with_approach(approach_pose: 'PoseStamped', goal_pose: 'PoseStamped', tip_link: str, root_link: str,
                                            robot_base_link: str) -> 'MoveResult':
    """
    Tries to achieve the goal_pose using the chain defined by tip_link and root_link. The approach_pose is used to drive
    the robot to a pose close the actual goal pose, the robot_base_link is used to define the base link of the robot.

    :param approach_pose: Pose near the goal_pose
    :param goal_pose: Pose to which the tip_link should be moved
    :param tip_link: The link which should be moved to goal_pose, usually the tool frame
    :param root_link: The start of the link chain which should be used for planning
    :param robot_base_link: The base link of the robot
    :return: A trajectory calculated to move the tip_link to the goal_pose
    """
    sync_worlds(projection=True)
    giskard_wrapper.allow_all_collisions()
    giskard_wrapper.set_cart_goal(approach_pose.ros_message(), robot_base_link, "map")
    giskard_wrapper.projection()
    giskard_wrapper.avoid_all_collisions()
    giskard_wrapper.set_cart_goal(goal_pose.ros_message(), tip_link, root_link)
    return giskard_wrapper.projection()


@init_giskard_interface
def projection_joint_goal(goal_poses: Dict[str, float], allow_collisions: bool = False) -> 'MoveResult':
    """
    Tries to achieve the joint goal defined by goal_poses, the goal_poses are projected to the closest point on the
    robot's workspace.

    :param goal_poses: Dictionary with joint names and position goals
    :param allow_collisions: If all collisions should be allowed for this goal
    :return: MoveResult message for this goal
    """
    sync_worlds(projection=True)
    if allow_collisions:
        giskard_wrapper.allow_all_collisions()
    giskard_wrapper.set_joint_goal(goal_poses)
    return giskard_wrapper.projection()


# Managing collisions

@init_giskard_interface
def allow_gripper_collision(gripper: Arms, at_goal: bool = False) -> None:
    """
    Allows the specified gripper to collide with anything.

    :param gripper: The gripper which can collide, either 'Arms.RIGHT', 'Arms.LEFT' or 'Arms.BOTH'
    :param at_goal: If the collision should be allowed only for this motion goal.
    """
    add_gripper_groups()
    for gripper_group in get_gripper_group_names():
        if gripper.name.lower() in gripper_group or gripper == Arms.BOTH:
            if at_goal:
                giskard_wrapper.motion_goals.allow_collision(gripper_group, CollisionEntry.ALL)
            else:
                giskard_wrapper.allow_collision(gripper_group, CollisionEntry.ALL)


@init_giskard_interface
def get_gripper_group_names() -> List[str]:
    """
    :return: The list of groups that are registered in giskard which have 'gripper' in their name.
    """
    groups = giskard_wrapper.get_group_names()
    return list(filter(lambda elem: "gripper" in elem, groups))


@init_giskard_interface
def add_gripper_groups() -> None:
    """
    Add the gripper links as a group for collision avoidance.

    :return: Response of the RegisterGroup Service
    """
    with giskard_lock:
        for name in giskard_wrapper.get_group_names():
            if "gripper" in name:
                return
        for description in RobotDescription.current_robot_description.get_manipulator_chains():
            giskard_wrapper.register_group(description.name + "_gripper", description.end_effector.start_link,
                                           RobotDescription.current_robot_description.name)


@init_giskard_interface
def avoid_all_collisions() -> None:
    """
    Will avoid all collision for the next goal.
    """
    giskard_wrapper.avoid_all_collisions()


@init_giskard_interface
def allow_self_collision() -> None:
    """
    Will allow the robot collision with itself.
    """
    giskard_wrapper.allow_self_collision()


@init_giskard_interface
def avoid_collisions(object1: Object, object2: Object) -> None:
    """
    Will avoid collision between the two objects for the next goal.

    :param object1: The first World Object
    :param object2: The second World Object
    """
    giskard_wrapper.avoid_collision(-1, object1.name, object2.name)


# Creating ROS messages

@init_giskard_interface
def make_world_body(object: Object) -> 'WorldBody':
    """
    Create a WorldBody message for a World Object. The WorldBody will contain the URDF of the World Object

    :param object: The World Object
    :return: A WorldBody message for the World Object
    """
    urdf_string = ""
    with open(object.path) as f:
        urdf_sting = f.read()
    urdf_body = WorldBody()
    urdf_body.type = WorldBody.URDF_BODY
    urdf_body.urdf = urdf_string

    return urdf_body


def make_point_stamped(point: List[float]) -> 'PointStamped':
    """
    Creates a PointStamped message for the given position in world coordinate frame.

    :param point: XYZ coordinates of the point
    :return: A PointStamped message
    """
    msg = PointStamped()
    msg.header.stamp = Time().now()
    msg.header.frame_id = "map"

    msg.point.x = point[0]
    msg.point.y = point[1]
    msg.point.z = point[2]

    return msg


def make_quaternion_stamped(quaternion: List[float]) -> 'QuaternionStamped':
    """
    Creates a QuaternionStamped message for the given quaternion.

    :param quaternion: The quaternion as a list of xyzw
    :return: A QuaternionStamped message
    """
    msg = QuaternionStamped()
    msg.header.stamp = Time().now()
    msg.header.frame_id = "map"

    msg.quaternion.x = quaternion[0]
    msg.quaternion.y = quaternion[1]
    msg.quaternion.z = quaternion[2]
    msg.quaternion.w = quaternion[3]

    return msg


def make_vector_stamped(vector: List[float]) -> 'Vector3Stamped':
    """
    Creates a Vector3Stamped message, this is similar to PointStamped but represents a vector instead of a point.

    :param vector: The vector given as xyz in world frame
    :return: A Vector3Stamped message
    """
    msg = Vector3Stamped()
    msg.header.stamp = Time().now()
    msg.header.frame_id = "map"

    msg.vector.x = vector[0]
    msg.vector.y = vector[1]
    msg.vector.z = vector[2]

    return msg



from .default_process_modules import *
from ..datastructures.enums import JointType
from ..datastructures.world import World
from ..designators.motion_designator import *
from ..external_interfaces.ik import request_giskard_ik
from ..external_interfaces.robokudo import *
from ..robot_description import RobotDescription


class StretchMoveHead(ProcessModule):
    """
    Process module for the simulated Stretch that moves the head such that it looks at the given position
    """

    def _execute(self, designator: MoveMotion) -> Any:
        target = designator.target
        robot = World.robot

        local_transformer = LocalTransformer()
        pose_in_pan = local_transformer.transform_pose(target, robot.get_link_tf_frame("link_head_pan"))

        new_pan = np.arctan2(pose_in_pan.position.y, pose_in_pan.position.x)
        current_pan = robot.get_joint_position("joint_head_pan")

        robot.set_joint_position("joint_head_pan", new_pan + current_pan)

        pose_in_tilt = local_transformer.transform_pose(target, robot.get_link_tf_frame("link_head_tilt"))
        new_tilt = np.arctan2(-pose_in_tilt.position.y,
                              np.sqrt(pose_in_tilt.position.z ** 2 + pose_in_tilt.position.x ** 2)) * -1
        current_tilt = robot.get_joint_position("joint_head_tilt")

        robot.set_joint_position("joint_head_tilt", new_tilt + current_tilt)


###########################################################
########## Process Modules for the Real Stretch ###########
###########################################################


class StretchNavigationReal(ProcessModule):
    """
    Process module for the real Stretch that sends a cartesian goal to giskard to move the robot base
    """

    def _execute(self, designator: MoveMotion) -> Any:
        logdebug(f"Sending goal to giskard to Move the robot")
        giskard.achieve_cartesian_goal(designator.target, RobotDescription.current_robot_description.base_link, "map")


class StretchMoveHeadReal(ProcessModule):
    """
    Process module for the real robot to move that such that it looks at the given position. Uses the same calculation
    as the simulated one
    """

    def _execute(self, desig: LookingMotion):
        target = desig.target
        robot = World.robot

        local_transformer = LocalTransformer()
        pose_in_pan = local_transformer.transform_pose(target, robot.get_link_tf_frame("head_pan_link"))

        new_pan = np.arctan2(pose_in_pan.position.y, pose_in_pan.position.x)

        current_pan = robot.get_joint_position("joint_head_pan")
        current_tilt = robot.get_joint_position("joint_head_tilt")

        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal({"head_pan_joint": new_pan + current_pan})

        pose_in_tilt = local_transformer.transform_pose(target, robot.get_link_tf_frame("head_tilt_link"))
        new_tilt = np.arctan2(-pose_in_tilt.position.y,
                              np.sqrt(pose_in_tilt.position.z ** 2 + pose_in_tilt.position.x ** 2)) * -1
        current_tilt = robot.get_joint_position("joint_head_tilt")
        giskard.avoid_all_collisions()
        giskard.achieve_joint_goal({"head_tilt_joint": new_tilt + current_tilt})


class StretchDetectingReal(ProcessModule):
    """
    Process Module for the real Stretch that tries to detect an object fitting the given object description. Uses Robokudo
    for perception of the environment.
    """

    def _execute(self, designator: DetectingMotion) -> Any:
        query_result = query(ObjectDesignatorDescription(types=[designator.object_type]))
        # print(query_result)
        obj_pose = query_result["ClusterPoseBBAnnotator"]

        lt = LocalTransformer()
        obj_pose = lt.transform_pose(obj_pose, World.robot.get_link_tf_frame("torso_lift_link"))
        obj_pose.orientation = [0, 0, 0, 1]
        obj_pose.position.x += 0.05

        bullet_obj = World.current_world.get_objects_by_type(designator.object_type)
        if bullet_obj:
            bullet_obj[0].set_pose(obj_pose)
            return bullet_obj[0]
        elif designator.object_type == ObjectType.JEROEN_CUP:
            cup = Object("cup", ObjectType.JEROEN_CUP, "jeroen_cup.stl", pose=obj_pose)
            return cup
        elif designator.object_type == ObjectType.BOWL:
            bowl = Object("bowl", ObjectType.BOWL, "bowl.stl", pose=obj_pose)
            return bowl

        return bullet_obj[0]


class StretchMoveGripperReal(ProcessModule):
    """
    Opens or closes the gripper of the real Stretch, gripper uses an action server for this instead of giskard
    """

    def _execute(self, designator: MoveGripperMotion) -> Any:
        chain = RobotDescription.current_robot_description.get_arm_chain(designator.gripper).get_static_gripper_state(
            designator.motion)
        giskard.achieve_joint_goal(chain)


class StretchManager(DefaultManager):
    def __init__(self):
        super().__init__()
        self.robot_name = "stretch_description"

    def navigate(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultNavigation(self._navigate_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return StretchNavigationReal(self._navigate_lock)

    def looking(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return StretchMoveHead(self._looking_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return StretchMoveHeadReal(self._looking_lock)

    def move_gripper(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveGripper(self._move_gripper_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return StretchMoveGripperReal(self._move_gripper_lock)

StretchManager()